"""Tauon Music Box - Module for DBus interaction"""

# Copyright © 2015-2019, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import logging
import threading
import time
import urllib.parse
from typing import TYPE_CHECKING

import dbus
import dbus.mainloop.glib
import dbus.service

from gi.repository import GLib

from tauon.t_modules.t_enums import PlayingState  # noqa: E402
from tauon.t_modules.t_extra import filename_to_metadata, star_count2  # noqa: E402

if TYPE_CHECKING:
	from tauon.t_modules.t_main import GuiVar, PlayerCtl, Tauon


class MPRIS(dbus.service.Object):
	def __init__(self, object_path: str, tauon: Tauon) -> None:
		self.tauon: Tauon = tauon
		self.gui: GuiVar = tauon.gui
		self.pctl: PlayerCtl = tauon.pctl
		bus = dbus.Bus(dbus.Bus.TYPE_SESSION)
		self.bus_name = dbus.service.BusName("org.mpris.MediaPlayer2.tauon", bus)  # This object must be kept alive

		# dbus.service.Object.__init__(self, bus_name, object_path)
		dbus.service.Object.__init__(self, bus, object_path, bus_name=self.bus_name)

		self.playing_index = -1
		self.last_playing_time: float = 0.0
		self.last_track_index: int = -1

		self.root_properties = {
			"CanQuit": True,
			#'Fullscreen'
			#'CanSetFullscreen'
			"CanRaise": True,
			"HasTrackList": False,
			"Identity": tauon.t_title,
			"DesktopEntry": tauon.t_id,
			"SupportedUriSchemes": dbus.Array([dbus.String("file")]),
			"SupportedMimeTypes": dbus.Array(
				[
					dbus.String("audio/mpeg"),
					dbus.String("audio/flac"),
					dbus.String("audio/ogg"),
					dbus.String("audio/m4a"),
				]
			),
		}

		self.player_properties = {
			"PlaybackStatus": "Stopped",
			"LoopStatus": self.get_loop_status(),
			"Rate": 1.0,
			"Shuffle": self.pctl.random_mode,
			"Volume": self.pctl.player_volume / 100,
			"Position": dbus.Int64(0),
			"MinimumRate": 1.0,
			"MaximumRate": 1.0,
			"CanGoNext": True,
			"CanGoPrevious": True,
			"CanPlay": True,
			"CanPause": True,
			"CanSeek": True,
			"CanControl": True,
			"Metadata": dbus.Dictionary({}, signature="sv"),
		}

	def update(self, force: bool = False) -> None:
		changed = {}

		if self.pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM):
			if self.player_properties["PlaybackStatus"] != "Playing":
				self.player_properties["PlaybackStatus"] = "Playing"
				changed["PlaybackStatus"] = self.player_properties["PlaybackStatus"]
		elif self.pctl.playing_state == PlayingState.STOPPED:
			if self.player_properties["PlaybackStatus"] != "Stopped":
				self.player_properties["PlaybackStatus"] = "Stopped"
				changed["PlaybackStatus"] = self.player_properties["PlaybackStatus"]
		elif self.pctl.playing_state == PlayingState.PAUSED:
			if self.player_properties["PlaybackStatus"] != "Paused":
				self.player_properties["PlaybackStatus"] = "Paused"
				changed["PlaybackStatus"] = self.player_properties["PlaybackStatus"]

		if self.pctl.player_volume / 100 != self.player_properties["Volume"]:
			self.player_properties["Volume"] = self.pctl.player_volume / 100
			changed["Volume"] = self.player_properties["Volume"]

		track = self.pctl.playing_object()
		if track is not None and (track.index != self.playing_index or force):
			self.playing_index = track.index
			id = f"/com/tauon/{track.index}/{abs(self.pctl.playlist_playing_position)}"
			if self.pctl.playing_state == PlayingState.URL_STREAM:
				id = "/com/tauon/radio"

			d = {
				"mpris:trackid": dbus.ObjectPath(id),
				"mpris:length": dbus.Int64(int(self.pctl.playing_length * 1000000)),
				"xesam:album": track.album,
				"xesam:albumArtist": dbus.Array([track.album_artist]),
				"xesam:artist": dbus.Array([track.artist]),
				"xesam:title": track.title,
				"xesam:asText": track.synced if track.synced else track.lyrics,
				"xesam:autoRating": star_count2(self.tauon.star_store.get(track.index)),
				"xesam:composer": dbus.Array([track.composer]),
				"tauon:loved": self.tauon.love(False, track.index),
				# added by msmafra
				"xesam:comment": dbus.Array([track.comment]),
				"xesam:genre": dbus.Array([track.genre]),
			}
			if not track.title:
				a, t = filename_to_metadata(track.filename)
				if not track.artist:
					d["xesam:artist"] = dbus.Array([a])
				d["xesam:title"] = t

			try:
				d["xesam:url"] = "file://" + urllib.parse.quote(track.fullpath)
			except Exception:
				logging.exception("Uri encode error")

			try:
				i_path = self.tauon.thumb_tracks.path(track)
				if i_path is not None:
					d["mpris:artUrl"] = "file://" + urllib.parse.quote(i_path)
			except Exception:
				logging.exception("Thumbnail error")
				logging.debug(track.fullpath.encode("utf-8", "replace").decode("utf-8"))

			self.update_progress()

			self.player_properties["Metadata"] = dbus.Dictionary(d, signature="sv")
			changed["Metadata"] = self.player_properties["Metadata"]

			if self.pctl.playing_state == PlayingState.URL_STREAM and self.player_properties["CanPause"] is True:
				self.player_properties["CanPause"] = False
				self.player_properties["CanSeek"] = False
				changed["CanPause"] = self.player_properties["CanPause"]
				changed["CanSeek"] = self.player_properties["CanSeek"]
			elif self.pctl.playing_state == PlayingState.PLAYING and self.player_properties["CanPause"] is False:
				self.player_properties["CanPause"] = True
				self.player_properties["CanSeek"] = True
				changed["CanPause"] = self.player_properties["CanPause"]
				changed["CanSeek"] = self.player_properties["CanSeek"]

		if len(changed) > 0:
			try:
				self.PropertiesChanged("org.mpris.MediaPlayer2.Player", changed, [])
			except Exception:
				logging.exception("Error updating MPRIS")
				logging.debug(changed)
				if track is not None:
					logging.debug(track.fullpath)

	def update_progress(self) -> None:
		if (
			self.pctl.repeat_mode
			and not self.pctl.album_repeat_mode
			and self.pctl.playing_object() is not None
			and self.last_track_index == self.pctl.playing_object().index
			and self.pctl.playing_time < self.last_playing_time
		):
			self.seek_do(self.pctl.playing_time)

		self.last_playing_time = self.pctl.playing_time
		self.last_track_index = self.pctl.playing_object().index

		self.player_properties["Position"] = dbus.Int64(int(self.pctl.playing_time * 1000000))

	def update_shuffle(self) -> None:
		self.player_properties["Shuffle"] = self.pctl.random_mode
		self.PropertiesChanged("org.mpris.MediaPlayer2.Player", {"Shuffle": self.pctl.random_mode}, [])

	def update_loop(self) -> None:
		self.player_properties["LoopStatus"] = self.get_loop_status()
		self.PropertiesChanged("org.mpris.MediaPlayer2.Player", {"LoopStatus": self.get_loop_status()}, [])

	def get_loop_status(self) -> str:
		if self.pctl.repeat_mode:
			if self.pctl.album_repeat_mode:
				return "Playlist"
			return "Track"
		return "None"

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2")
	def Raise(self) -> None:
		self.tauon.request_raise()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2")
	def Quit(self) -> None:
		self.tauon.wake()
		self.tauon.exit("Exit request received from MPRIS2")

	@dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature="ss", out_signature="v")
	def Get(self, interface_name: str, property_name: str) -> dict | None:
		if interface_name == "org.mpris.MediaPlayer2":
			# return self.GetAll(interface_name)[property_name]
			return self.root_properties[property_name]
		if interface_name == "org.mpris.MediaPlayer2.Player":
			return self.player_properties[property_name]
		return None

	@dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature="s", out_signature="a{sv}")
	def GetAll(self, interface_name: str) -> dict:
		if interface_name == "org.mpris.MediaPlayer2":
			return self.root_properties
		if interface_name == "org.mpris.MediaPlayer2.Player":
			return self.player_properties
		return {}

	@dbus.service.method(dbus_interface=dbus.PROPERTIES_IFACE, in_signature="ssv", out_signature="")
	def Set(self, interface_name: str, property_name: str, value: str) -> None:
		if interface_name == "org.mpris.MediaPlayer2.Player":
			self.tauon.wake()
			if property_name == "Volume":
				self.pctl.player_volume = min(max(int(value * 100), 0), 100)
				self.pctl.set_volume()
				self.gui.update += 1
			if property_name == "Shuffle":
				self.pctl.random_mode = bool(value)
				self.update_shuffle()
				self.gui.update += 1
			if property_name == "LoopStatus":
				if value == "None":
					self.tauon.menu_repeat_off()
				elif value == "Track":
					self.tauon.menu_set_repeat()
				elif value == "Playlist":
					self.tauon.menu_album_repeat()
				self.gui.update += 1

		if interface_name == "org.mpris.MediaPlayer2":
			pass

	@dbus.service.signal(dbus_interface=dbus.PROPERTIES_IFACE, signature="sa{sv}as")
	def PropertiesChanged(self, interface_name: str, change: dict, inval: list) -> None:
		pass

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Next(self) -> None:
		self.tauon.wake()
		self.pctl.advance()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Previous(self) -> None:
		self.tauon.wake()
		self.pctl.back()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Pause(self) -> None:
		self.pctl.pause_only()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def PlayPause(self) -> None:
		self.tauon.wake()
		if self.pctl.playing_state == PlayingState.URL_STREAM:
			self.pctl.stop()  # Stop if playing radio
		else:
			self.pctl.play_pause()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Stop(self) -> None:
		self.pctl.stop()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Play(self) -> None:
		self.tauon.wake()
		self.pctl.play()

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Seek(self, offset: int) -> None:
		self.pctl.seek_time(self.pctl.playing_time + (offset / 1000000))

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def SetPosition(self, id: str, position: int) -> None:
		self.pctl.seek_time(position / 1000000)

		self.player_properties["Position"] = dbus.Int64(int(position))
		self.Seeked(self.pctl.playing_time)

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def OpenUri(self, uri: str) -> None:
		self.tauon.wake()
		self.tauon.open_uri(uri)

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def LovePlaying(self) -> None:
		if not self.tauon.love(set=False):
			self.tauon.love(set=True, no_delay=True)
			self.update(True)
			self.gui.pl_update += 1

	@dbus.service.method(dbus_interface="org.mpris.MediaPlayer2.Player")
	def UnLovePlaying(self) -> None:
		if self.tauon.love(set=False):
			self.tauon.love(set=True, no_delay=True)
			self.update(True)
			self.gui.pl_update += 1

	@dbus.service.signal(dbus_interface="org.mpris.MediaPlayer2.Player")
	def Seeked(self, position: str) -> None:
		pass

	def seek_do(self, seconds: str | float) -> None:
		self.Seeked(dbus.Int64(int(seconds * 1000000)))


class Gnome:
	def __init__(self, tauon: Tauon) -> None:
		self.bus_object = None
		self.tauon: Tauon = tauon
		self.resume_playback: bool = False

	def focus(self) -> None:
		if self.bus_object is not None:
			try:
				# this is what gives us the multi media keys.
				dbus_interface = "org.gnome.SettingsDaemon.MediaKeys"
				self.bus_object.GrabMediaPlayerKeys("TauonMusicBox", 0, dbus_interface=dbus_interface)
			except Exception:
				logging.exception("Error connecting to org.gnome.SettingsDaemon.MediaKeys")

	def main(self) -> None:
		prefs = self.tauon.prefs
		gui = self.tauon.gui
		pctl = self.tauon.pctl
		tauon = self.tauon

		def on_mediakey(comes_from: str, what: str) -> None:
			if what == "Play":
				self.tauon.inp.media_key = "Play"
			elif what == "Pause":
				self.tauon.inp.media_key = "Pause"
			elif what == "Stop":
				self.tauon.inp.media_key = "Stop"
			elif what == "Next":
				self.tauon.inp.media_key = "Next"
			elif what == "Previous":
				self.tauon.inp.media_key = "Previous"
			elif what == "Rewind":
				self.tauon.inp.media_key = "Rewind"
			elif what == "FastForward":
				self.tauon.inp.media_key = "FastForward"
			elif what == "Repeat":
				self.tauon.inp.media_key = "Repeat"
			elif what == "Shuffle":
				self.tauon.inp.media_key = "Shuffle"

			if self.tauon.inp.media_key:
				gui.update = 1

		try:
			# set up the glib main loop.
			dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

			bus = dbus.Bus(dbus.Bus.TYPE_SYSTEM)
			bus_object = bus.get_object("org.freedesktop.login1", "/org/freedesktop/login1")

			iface = dbus.Interface(bus_object, dbus_interface="org.freedesktop.login1.Manager")

			tauon.sleep_lock = iface.Inhibit("sleep", "Tauon Music Box", "Pause music on sleep", "delay")
			tauon.shutdown_lock = iface.Inhibit("shutdown", "Tauon Music Box", "Save data to disk", "delay")

			def update_play_lock() -> None:
				if prefs.block_suspend:
					if (
						pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM)
						and tauon.play_lock is None
					):
						tauon.play_lock = iface.Inhibit("idle", "Tauon Music Box", "Audio is playing", "block")
					elif (
						pctl.playing_state not in (PlayingState.PLAYING, PlayingState.URL_STREAM)
						and tauon.play_lock is not None
					):
						del tauon.play_lock
						tauon.play_lock = None
				elif tauon.play_lock is not None:
					del tauon.play_lock
					tauon.play_lock = None

			tauon.update_play_lock = update_play_lock

			def PrepareForSleep(active: int) -> None:
				if active == 1 and tauon.sleep_lock is not None:
					logging.info("System is suspending!")
					if pctl.playing_state == PlayingState.URL_STREAM:
						pctl.stop(block=True)
						if prefs.resume_play_wake:
							pctl.playing_state = PlayingState.URL_STREAM
							self.resume_playback = True
					elif pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM):
						tauon.pctl.pause()
						if prefs.resume_play_wake:
							self.resume_playback = True
					del tauon.sleep_lock
					tauon.sleep_lock = None

				elif active == 0 and tauon.sleep_lock is None:
					tauon.sleep_lock = iface.Inhibit("sleep", "Tauon Music Box", "Pause music on sleep", "delay")
					if self.resume_playback:
						self.resume_playback = False
						if pctl.playing_state == PlayingState.URL_STREAM:
							pctl.playing_state = PlayingState.STOPPED
							time.sleep(4)
							pctl.play()
							logging.info("Resume Radio")
						else:
							pctl.play()

			def PrepareForShutdown(active: int) -> None:
				if active != 1:
					return
				logging.info("The system is shutting down!")
				tauon.quick_close = True
				tauon.exit("System shutdown signal received")

			iface.connect_to_signal("PrepareForSleep", PrepareForSleep)
			iface.connect_to_signal("PrepareForShutdown", PrepareForShutdown)

		except Exception:
			logging.exception("Failure to connect to login1")

		# t_bus = dbus.Bus(dbus.Bus.TYPE_SESSION)
		# t_bus_name = dbus.service.BusName('com.github.taiko2k.tauonmb', t_bus)  # This object must be kept alive
		#
		# class T(dbus.service.Object):
		# @dbus.service.method(
		# "com.github.taiko2k.tauonmb",
		# in_signature='a{sv}', out_signature='')
		# def start(self, options={}) -> None:
		# logging.info("START")
		#
		# def __init__(self, object_path) -> None:
		# dbus.service.Object.__init__(self, t_bus, object_path, bus_name=t_bus_name)
		#
		# pctl.sgl = T("/")

		# ----------
		if prefs.enable_mpris:
			try:
				pctl.mpris = MPRIS("/org/mpris/MediaPlayer2", tauon)

			except Exception:
				logging.exception("MPRIS2 CONNECT FAILED")

		mainloop = GLib.MainLoop()
		mainloop.run()



"""Tauon Music Box - Phazor audio backend module"""

# Copyright © 2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import ctypes
import logging
import sys
from http import HTTPStatus

if sys.platform == "win32":
	from ctypes import WINFUNCTYPE
import hashlib
import importlib.machinery
import math
import os.path
import shutil
import subprocess
import sysconfig
import tempfile
import threading
import time
from ctypes import CFUNCTYPE, POINTER, c_char, c_char_p, c_int, c_void_p, cast
from pathlib import Path
from typing import TYPE_CHECKING

import requests
from requests.models import PreparedRequest

from tauon.t_modules.t_enums import PlayerState, PlayingState
from tauon.t_modules.t_extra import Timer, shooter, tmp_cache_dir

if TYPE_CHECKING:
	from ctypes import CDLL

	from tauon.t_modules.t_main import GuiVar, PlayerCtl, Tauon, TrackClass
	from tauon.t_modules.t_prefs import Prefs


class FFRun:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon: Tauon = tauon
		self.decoder = None

	def close(self) -> None:
		if self.decoder:
			self.decoder.terminate()
			if self.decoder.stdin:
				logging.debug("Closing STDIN in FFrun")
				self.decoder.stdin.close()
			if self.decoder.stdout:
				logging.debug("Closing STDOUT in FFrun")
				self.decoder.stdout.close()
			if self.decoder.stderr:
				logging.debug("Closing STDERR in FFrun")
				self.decoder.stderr.close()
			self.decoder.wait()  # Ensure the process fully terminates
		self.decoder = None

	def start(self, uri: bytes, start_ms: int, samplerate: int) -> int:
		self.close()
		ffmpeg_path = self.tauon.get_ffmpeg()
		if ffmpeg_path is None:
			self.tauon.test_ffmpeg()
			return 1
		path = str(ffmpeg_path)
		cmd = [path]
		cmd += ["-loglevel", "quiet"]
		if start_ms > 0:
			cmd += ["-ss", f"{start_ms}ms"]
		cmd += ["-i", uri.decode(), "-acodec", "pcm_s16le", "-f", "s16le", "-ac", "2", "-ar", f"{samplerate}", "-"]
		startupinfo = None
		if sys.platform == "win32":
			startupinfo = subprocess.STARTUPINFO()
			startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
		try:
			self.decoder = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, startupinfo=startupinfo)
		except Exception:
			logging.exception("Failed to start FFmpeg")
			return 1
		return 0

	def read(self, buffer: int, maximum: int) -> int:
		if self.decoder:
			data = self.decoder.stdout.read(maximum)
			p = cast(buffer, POINTER(c_char * maximum))
			p.contents.value = data
			return len(data)
		return 0


class StreamFeeder:
	"""Feeds network audio data into the PHAzOR in-memory stream buffer.

	Implements the producer half of phazor's network byte stream. Polls the
	library for the wanted byte offset, performs HTTP range requests and
	pushes data in as it becomes available. When the decoder seeks outside
	the buffered window, phazor requests a new offset and the feeder
	responds with a new range request.
	"""

	def __init__(self, tauon: Tauon, aud: CDLL) -> None:
		self.tauon = tauon
		self.pctl = tauon.pctl
		self.aud = aud
		self.session = requests.Session()
		self.enabled = True
		self.speed = 0.0  # measured download rate in bytes/sec, 0 when not downloading
		self._speed_bytes = 0
		self._speed_t0 = time.monotonic()
		# Pending request to also save a fully-streamed track to the disk cache
		self._cache_lock = threading.Lock()
		self._cache_request: tuple[TrackClass, str] | None = None
		# Prefetch of the upcoming track's first few MB
		self.prefetch_bytes = 3 * 1000 * 1000
		self._prefetch_session = requests.Session()  # separate from the main loop's
		self._prefetch_lock = threading.Lock()
		self._prefetch_target: TrackClass | None = None  # track to prefetch
		self._prefetch_url: str | None = None  # url of the held prefetch data
		self._prefetch_data: bytes = b""
		self._prefetch_size: int = -1  # full file size if known
		self._prefetch_done: bool = False
		self._prefetching: bool = False
		try:
			aud.net_generation.restype = ctypes.c_int
			aud.net_get_url.restype = ctypes.c_char_p
			aud.net_want.restype = ctypes.c_longlong
			aud.net_want.argtypes = (ctypes.c_int,)
			aud.net_feed.restype = ctypes.c_int
			aud.net_feed.argtypes = (ctypes.c_int, ctypes.c_longlong, ctypes.c_char_p, ctypes.c_int)
			aud.net_set_size.restype = None
			aud.net_set_size.argtypes = (ctypes.c_int, ctypes.c_longlong)
			aud.net_eof.restype = None
			aud.net_eof.argtypes = (ctypes.c_int,)
			aud.net_fail.restype = None
			aud.net_fail.argtypes = (ctypes.c_int,)
			aud.net_set_seekable.restype = None
			aud.net_set_seekable.argtypes = (ctypes.c_int, ctypes.c_int)
			aud.set_load_net.restype = None
			aud.set_load_net.argtypes = (ctypes.c_int,)
		except AttributeError:
			logging.warning("This PHAzOR build does not support network streaming")
			self.enabled = False
			return
		shoot = threading.Thread(target=self._run)
		shoot.daemon = True
		shoot.start()

	def _speed_tick(self, n: int) -> None:
		"""Account n downloaded bytes towards the measured rate"""
		now = time.monotonic()
		self._speed_bytes += n
		elapsed = now - self._speed_t0
		if elapsed >= 0.5:
			self.speed = self._speed_bytes / elapsed
			self._speed_bytes = 0
			self._speed_t0 = now

	def _speed_idle(self) -> None:
		"""Not downloading; the rate is meaningless right now"""
		self.speed = 0.0
		self._speed_bytes = 0
		self._speed_t0 = time.monotonic()

	def request_cache(self, track: TrackClass, url: str) -> None:
		"""Ask the feeder to also save this stream to the disk cache if it
		downloads fully and contiguously. Matched to the stream by URL."""
		with self._cache_lock:
			self._cache_request = (track, url)

	def set_prefetch(self, track: TrackClass | None) -> None:
		"""Set the upcoming track whose start should be prefetched once the
		current stream finishes downloading. Passing a different track (or
		None) discards any previously prefetched data."""
		with self._prefetch_lock:
			if track is self._prefetch_target:
				return
			self._prefetch_target = track
			self._prefetch_url = None
			self._prefetch_data = b""
			self._prefetch_size = -1
			self._prefetch_done = False

	def _resolve_stream_url(self, track: TrackClass) -> str | None:
		"""Resolve a track to a single streamable http(s) URL, or None"""
		try:
			network_url, params = self.pctl.get_url(track)
		except Exception:
			logging.exception("Prefetch: failed to resolve URL")
			return None
		if not isinstance(network_url, str) or not network_url.startswith("http"):
			return None
		if params:
			req = PreparedRequest()
			req.prepare_url(network_url, params)
			return req.url
		return network_url

	def _prefetch_worker(self, track: TrackClass) -> None:
		"""Fetch the first prefetch_bytes of the upcoming track into memory"""
		title = getattr(track, "title", None) or getattr(track, "filename", "?")
		url = None
		data = b""
		size = -1
		try:
			# Skip if it is already fully cached on disk
			cachement = self.tauon.cachement
			if cachement is not None and cachement.get_local_instant(track) is not None:
				logging.info(f"Prefetch: skip '{title}' (already cached)")
				return
			url = self._resolve_stream_url(track)
			if not url:
				logging.info(f"Prefetch: skip '{title}' (no streamable URL)")
				return
			logging.info(f"Prefetch: start '{title}' (up to {self.prefetch_bytes // 1000} KB)")
			resp = self._prefetch_session.get(
				url, headers={"Range": f"bytes=0-{self.prefetch_bytes - 1}"}, stream=True, timeout=(5, 30))
			try:
				if resp.status_code != HTTPStatus.PARTIAL_CONTENT:
					# Without range support we cannot resume after the prefetch,
					# so prefetching would not help
					logging.info(f"Prefetch: skip '{title}' (no range support, status {resp.status_code})")
					return
				total = resp.headers.get("Content-Range", "").rpartition("/")[2]
				if total.isdigit():
					size = int(total)
				buf = bytearray()
				for chunk in resp.iter_content(65536):
					# Abandon if the target changed (queue edit / skip)
					with self._prefetch_lock:
						if self._prefetch_target is not track:
							logging.info(f"Prefetch: abort '{title}' (target changed) after {len(buf) // 1000} KB")
							return
					buf += chunk
					if len(buf) >= self.prefetch_bytes:
						break
				data = bytes(buf)
			finally:
				resp.close()
		except Exception:
			logging.exception(f"Prefetch: download failed for '{title}'")
			return
		finally:
			with self._prefetch_lock:
				if self._prefetch_target is track and url and data:
					self._prefetch_url = url
					self._prefetch_data = data
					self._prefetch_size = size
					self._prefetch_done = True
					logging.info(f"Prefetch: ready '{title}' ({len(data) // 1000} KB of {size // 1000 if size > 0 else '?'} KB)")
				self._prefetching = False

	def _maybe_start_prefetch(self) -> None:
		"""Kick off a background prefetch of the upcoming track if pending"""
		with self._prefetch_lock:
			if self._prefetching or self._prefetch_done:
				return
			track = self._prefetch_target
			if track is None or not getattr(track, "is_network", False):
				return
			self._prefetching = True
		shoot = threading.Thread(target=self._prefetch_worker, args=(track,))
		shoot.daemon = True
		shoot.start()

	def _cache_open(self, track: TrackClass):  # noqa: ANN202 - returns (file, temp, final) or None
		"""Open a temp file in the cache dir for teeing a stream to disk"""
		try:
			cachement = self.tauon.cachement
			if cachement is None:
				return None
			key = cachement.get_key(track)
			if key in cachement.files:
				return None  # already cached
			final_path = os.path.join(cachement.direc, key)
			if os.path.isfile(final_path):
				return None
			fd, temp = tempfile.mkstemp(dir=cachement.direc, suffix=".part")
			return os.fdopen(fd, "wb"), temp, final_path
		except Exception:
			logging.exception("Stream cache: could not open temp file")
			return None

	def _cache_commit(self, track: TrackClass, temp_path: str, final_path: str, size: int) -> None:
		"""Move a completed temp file into the cache and register it with Cachement"""
		try:
			cachement = self.tauon.cachement
			os.replace(temp_path, final_path)
			key = os.path.basename(final_path)
			if key not in cachement.files:
				cachement.files.append(key)
			if key not in cachement.list:
				cachement.list.append(key)
			cachement.trim_cache()
			logging.info(f"Cached streamed track to disk ({size} bytes)")
		except Exception:
			logging.exception("Stream cache: finalize failed")
			try:
				os.remove(temp_path)
			except OSError:
				pass

	def _run(self) -> None:
		aud = self.aud
		gen = -1
		resp = None
		offset = 0
		size = -1
		pending = b""
		no_range = False
		retries = 0

		# Disk-cache tee state for the current stream
		cache_file = None
		cache_temp = None
		cache_final = None
		cache_track = None
		cache_offset = 0  # contiguous bytes written from offset 0
		cache_ok = False

		def close_response() -> None:
			nonlocal resp, pending
			if resp is not None:
				try:
					resp.close()
				except Exception:
					logging.exception("Stream: error closing response")
				resp = None
			pending = b""

		def cache_reset() -> None:
			nonlocal cache_file, cache_temp, cache_final, cache_track, cache_offset, cache_ok
			cache_file = None
			cache_temp = None
			cache_final = None
			cache_track = None
			cache_offset = 0
			cache_ok = False

		def cache_discard() -> None:
			# Abandon an incomplete tee and delete its temp file
			nonlocal cache_file, cache_temp
			if cache_file is not None:
				try:
					cache_file.close()
				except Exception:
					logging.exception("Stream cache: close error")
				if cache_temp:
					try:
						os.remove(cache_temp)
					except OSError:
						pass
			cache_reset()

		def cache_commit() -> None:
			# The whole file was streamed contiguously; persist it
			nonlocal cache_file
			if cache_file is None:
				return
			try:
				cache_file.close()
			except Exception:
				logging.exception("Stream cache: close error")
			self._cache_commit(cache_track, cache_temp, cache_final, cache_offset)
			cache_reset()

		while True:
			try:
				g = aud.net_generation()
				if g == -1:
					close_response()
					cache_discard()
					gen = -1
					self._speed_idle()
					time.sleep(0.04)
					continue
				if g != gen:
					# A new stream was opened
					close_response()
					cache_discard()
					gen = g
					size = -1
					no_range = False
					retries = 0
					# Adopt a pending cache request if it matches this stream
					req = None
					with self._cache_lock:
						if self._cache_request is not None:
							req = self._cache_request
							self._cache_request = None
					try:
						url0 = aud.net_get_url().decode()
					except Exception:
						url0 = ""
					if req is not None and req[1] == url0:
						opened = self._cache_open(req[0])
						if opened is not None:
							cache_file, cache_temp, cache_final = opened
							cache_track = req[0]
							cache_offset = 0
							cache_ok = True

					# Inject prefetched bytes for this stream, if we have them
					inj = None
					with self._prefetch_lock:
						if self._prefetch_done and self._prefetch_data and self._prefetch_url == url0:
							inj_title = getattr(self._prefetch_target, "title", None) or "?"
							inj = (self._prefetch_data, self._prefetch_size)
							self._prefetch_url = None
							self._prefetch_data = b""
							self._prefetch_done = False
							self._prefetch_target = None
					if inj is not None:
						pdata, psize = inj
						if psize > 0:
							aud.net_set_size(gen, psize)
							size = psize
						fpos = 0
						while fpos < len(pdata):
							fed0 = aud.net_feed(gen, fpos, pdata[fpos:], len(pdata) - fpos)
							if fed0 <= 0:
								break
							if cache_file is not None and cache_ok and fpos == cache_offset:
								try:
									cache_file.write(pdata[fpos:fpos + fed0])
									cache_offset += fed0
								except Exception:
									logging.exception("Stream cache: prefetch write error")
									cache_discard()
							fpos += fed0
						offset = fpos
						if offset > 0:
							logging.info(f"Prefetch: used '{inj_title}' (injected {offset // 1000} KB into the stream)")

				want = aud.net_want(gen)
				if want == -2:
					close_response()
					continue
				if want == -1:
					# Buffered through to end of file. If the whole file came
					# down contiguously, persist it to the disk cache.
					close_response()
					self._speed_idle()
					if cache_file is not None and cache_ok and size > 0 and cache_offset >= size:
						cache_commit()
					# Current track is fully buffered; prefetch the next one
					self._maybe_start_prefetch()
					time.sleep(0.05)
					continue

				if pending and offset == want:
					pass  # previous chunk is still waiting to be accepted
				elif resp is None or offset != want:
					# (Re)open the request at the wanted offset
					close_response()
					if no_range and want > 0:
						logging.error("Stream: server does not support range requests")
						aud.net_fail(gen)
						continue
					url = aud.net_get_url().decode()
					try:
						resp = self.session.get(
							url, headers={"Range": f"bytes={want}-"}, stream=True, timeout=(5, 30))
					except Exception:
						logging.exception("Stream: request failed")
						retries += 1
						if retries > 3:
							aud.net_fail(gen)
							retries = 0
						else:
							time.sleep(0.5)
						continue
					code = resp.status_code
					if code == HTTPStatus.PARTIAL_CONTENT:
						total = resp.headers.get("Content-Range", "").rpartition("/")[2]
						if total.isdigit():
							size = int(total)
							aud.net_set_size(gen, size)
					elif code == HTTPStatus.OK and want == 0:
						# Server ignored the range header; play linearly
						no_range = True
						aud.net_set_seekable(gen, 0)
						cl = resp.headers.get("Content-Length", "")
						if cl.isdigit():
							size = int(cl)
							aud.net_set_size(gen, size)
					else:
						logging.error(f"Stream: server returned status {code}")
						close_response()
						aud.net_fail(gen)
						continue
					offset = want
					# Don't tee files larger than the cache limit to disk
					if cache_file is not None and size > 0:
						limit_b = max(self.tauon.prefs.cache_limit, 0) * 1000 * 1000
						if limit_b and size > limit_b:
							cache_discard()

				if not pending:
					try:
						chunk = resp.raw.read(65536)
					except Exception:
						logging.exception("Stream: read failed")
						close_response()
						retries += 1
						if retries > 3:
							aud.net_fail(gen)
							retries = 0
						time.sleep(0.2)
						continue
					if not chunk:
						# End of response body
						close_response()
						if size < 0 or offset >= size:
							aud.net_eof(gen)
							# Whole file streamed contiguously, persist it
							if cache_file is not None and cache_ok and cache_offset == offset:
								cache_commit()
							else:
								cache_discard()
						else:
							# Connection ended early, resume from current offset
							retries += 1
							if retries > 3:
								aud.net_fail(gen)
								retries = 0
							else:
								time.sleep(0.2)
						continue
					retries = 0
					pending = chunk

				fed = aud.net_feed(gen, offset, pending, len(pending))
				if fed in (-1, -2):
					# Stream gone, or a seek redirected the wanted offset
					close_response()
					continue
				if fed == 0:
					# Buffer is full, wait for the player to consume some
					self._speed_tick(0)
					time.sleep(0.05)
					continue
				# Tee contiguous bytes to the disk cache, if enabled. A seek
				# (offset no longer matching the write head) makes a complete
				# file impossible, so the tee is abandoned.
				if cache_file is not None and cache_ok:
					if offset == cache_offset:
						try:
							cache_file.write(pending[:fed])
							cache_offset += fed
						except Exception:
							logging.exception("Stream cache: write error")
							cache_discard()
					else:
						cache_discard()
				offset += fed
				self._speed_tick(fed)
				pending = pending[fed:] if fed < len(pending) else b""
				if size > 0:
					self.pctl.buffering_percent = min(100, int(offset * 100 / size))
			except Exception:
				logging.exception("Stream feeder error")
				close_response()
				time.sleep(0.5)


class Cachement:
	def __init__(self, tauon: Tauon) -> None:
		# fmt:off
		self.tauon        = tauon
		self.gui          = tauon.gui
		self.pctl         = tauon.pctl
		self.prefs        = tauon.prefs
		self.show_message = tauon.show_message
		self.audio_cache  = tauon.cache_directory / "network-audio1"
		self.audio_cache2 = tauon.cache_directory / "audio-cache"
		# fmt:on
		self.direc = str(self.audio_cache2)
		if self.prefs.tmp_cache:
			self.direc = os.path.join(tmp_cache_dir(), "audio-cache")
		if not Path(self.direc).exists():
			os.makedirs(self.direc)
		self.list: list[str] = self.prefs.cache_list
		self.files = os.listdir(self.direc)
		self.get_now = None
		self.running = False
		self.ready = None
		self.error = None

	def get_key(self, track: TrackClass) -> str:
		if track.is_network:
			return hashlib.sha256((str(track.index) + track.url_key).encode()).hexdigest()
		return hashlib.sha256(track.fullpath.encode()).hexdigest()

	def get_file_cached_only(self, track: TrackClass) -> str | None:
		key = self.get_key(track)
		if key in self.files:
			path = Path(self.direc) / key
			if path.is_file():
				return str(path)
		return None

	def get_local_instant(self, track: TrackClass) -> str | None:
		"""Return a complete cached file or existing transcode for this track, if any"""
		path = self.get_file_cached_only(track)
		if path:
			return path
		for codec in (".opus", ".ogg", ".flac", ".mp3"):
			idea = (
				os.path.join(
					self.prefs.encoder_output, self.tauon.encode_folder_name(track), self.tauon.encode_track_name(track)
				)
				+ codec
			)
			if os.path.isfile(idea):
				return idea
		return None

	def get_file(self, track: TrackClass) -> tuple[int, str | None]:
		# 0: file ready
		# 1: file downloading
		# 2: file not found
		if self.error == track:
			return 2, None

		key = self.get_key(track)
		path = os.path.join(self.direc, key)

		if self.running and self.get_now == track:
			return 1, path

		if key in self.files and os.path.isfile(path):
			logging.info("Got cached file")
			self.files.remove(key)
			self.files.append(key)  # bump to top of list
			self.get_now = None
			if not self.running:
				shoot_dl = threading.Thread(target=self.run)
				shoot_dl.daemon = True
				shoot_dl.start()
			return 0, path

		# disable me for debugging
		for codec in (".opus", ".ogg", ".flac", ".mp3"):
			idea = (
				os.path.join(
					self.prefs.encoder_output, self.tauon.encode_folder_name(track), self.tauon.encode_track_name(track)
				)
				+ codec
			)
			if os.path.isfile(idea):
				logging.info("Found transcode")
				return 0, idea

		self.get_now = track
		if not self.running:
			shoot_dl = threading.Thread(target=self.run)
			shoot_dl.daemon = True
			shoot_dl.start()
		return 1, path

	def run(self) -> None:
		self.running = True

		now = self.get_now
		self.get_now = None
		if now is not None:
			error = self.dl_file(now)
			if error:
				self.error = now
				self.running = False
				return

		if self.get_now is None:
			i = 0
			while i < 10:
				time.sleep(0.1)
				i += 1
				if self.get_now is not None:
					self.running = False
					return
			logging.info("Precache next track")
			next_track = self.pctl.advance(dry=True)
			if next_track is not None:
				self.dl_file(self.pctl.get_track(next_track))

		self.trim_cache()
		self.running = False
		return

	def trim_cache(self) -> None:
		# Remove untracked items
		for item in self.files:
			t = os.path.join(self.direc, item)
			if os.path.isfile(t) and item not in self.list:
				os.remove(t)

		# Check total size
		limit = self.prefs.cache_limit
		if self.prefs.tmp_cache:
			limit = 10
		while True:
			s = 0
			for item in list(self.list):
				t = os.path.join(self.direc, item)
				if not os.path.exists(t):
					self.list.remove(item)
					continue
				s += os.path.getsize(t)
			# Removed oldest items if over limit
			if s > limit * 1000 * 1000 and len(self.list) > 3:
				t = os.path.join(self.direc, self.list[0])
				os.remove(t)
				del self.list[0]
			else:
				break

	def dl_file(self, track: TrackClass) -> int | None:
		self.pctl.buffering_percent = 0
		key = self.get_key(track)
		path = Path(self.direc) / key
		if path.exists():
			if not path.is_file():
				return 1
			if key in self.list:
				return 0
			path.unlink()
		if not track.is_network:
			if not Path(track.fullpath).is_file():
				self.error = track
				self.running = False
				return 1

			logging.info("Start transfer")
			timer = Timer()
			target = path.open("wb")
			source = Path(track.fullpath).open("rb")
			while True:
				try:
					data = source.read(128000)
				except Exception:
					logging.exception("Transfer failed.")
					break
				if len(data) > 0:
					logging.info(f"Caching file @ {int(len(data) / timer.hit() / 1000)} kbps")
				else:
					break
				target.write(data)
			target.close()
			source.close()
			logging.info("got file")
			self.files.append(key)
			self.list.append(key)
			return 0

		try:
			logging.info("Download file")

			network_url, params = self.pctl.get_url(track)
			if not network_url:
				logging.info("No URL")
				return 1
			if type(network_url) in (list, tuple) and len(network_url) == 1:
				network_url = network_url[0]
			elif type(network_url) in (list, tuple):
				logging.info("Multi part DL")
				logging.info(path)
				# fmt:off
				ffmpeg_command = [
					str(self.tauon.get_ffmpeg()),
					"-i", "-",      # Input from stdin (pipe the data)
					"-f", "flac",   # Specify FLAC as the output format explicitly
					"-c:a", "copy", # Copy FLAC data without re-encoding
					path,           # Output file for extracted FLAC
				]
				# fmt:on
				p = subprocess.Popen(ffmpeg_command, stdin=subprocess.PIPE)
				i = 0
				for url in network_url:
					i += 1
					logging.info(i)
					response = requests.get(url, timeout=10)
					if response.status_code == HTTPStatus.OK:
						p.stdin.write(response.content)
					else:
						logging.error(f"ERROR CODE: {response.status_code}")
					if i == 3:
						self.ready = track

					self.gui.update += 1
					self.pctl.buffering_percent = math.floor(i / len(network_url) * 100)
					self.gui.buffering_text = str(self.pctl.buffering_percent) + "%"
					if self.get_now is not None and self.get_now != track:
						logging.info("Aborted loading track!")
						return None

				# logging.info("done")
				p.stdin.close()
				p.wait()

				logging.info("Done loading track")
				# self.ready = track
				self.files.append(key)
				self.list.append(key)
				return None

			part = requests.get(network_url, stream=True, params=params, timeout=(3, 10))

			if part.status_code == HTTPStatus.NOT_FOUND:
				self.show_message("Server: File not found", mode="error")
				self.error = track
				return 1
			if part.status_code != HTTPStatus.OK:
				self.show_message("Server Error", mode="error")
				self.error = track
				return 1

		except Exception as e:
			logging.exception("Download failed!")
			self.show_message(_("Error"), str(e), mode="error")
			self.error = track
			return 1

		a = 0
		length = 0
		cl = part.headers.get("Content-Length")
		if cl:
			length = int(cl)
			self.gui.buffering_text = "0%"

		timer = Timer()
		try:
			with path.open("wb") as f:
				for chunk in part.iter_content(chunk_size=1024):
					if chunk:  # filter out keep-alive new chunks
						a += 1
						if a == 3000 and self.prefs.jump_start:  # kilobyes~
							self.ready = track
						if a % 32 == 0:
							# time.sleep(0.03)
							# logging.info(f"Downloading file @ {round(32 / timer.hit())} kbps")
							if length:
								self.gui.update += 1
								self.pctl.buffering_percent = round(a * 1000 / length * 100)
								self.gui.buffering_text = str(round(a * 1000 / length * 100)) + "%"

						if self.get_now is not None and self.get_now != track:
							logging.warning("Aborted loading track!")
							return None

						# if self.cancel is True:
						# 	self.part.close()
						# 	self.status = "failed"
						# 	logging.info("Abort download")
						# 	return

						f.write(chunk)
			logging.info("Done loading track")
			self.ready = track
			self.files.append(key)
			self.list.append(key)
		except Exception:
			logging.exception("Download failed!")
			return 1
		return 0


def find_library(libname: str) -> Path | None:
	"""Search for 'libname.extension' in various formats.

	Return library Path loadable with ctypes.CDLL, otherwise return None
	"""
	# Can look like ~/Projects/Tauon/t_modules
	base_path = Path(__file__).parent
	search_paths: list[Path] = []
	so_extensions = importlib.machinery.EXTENSION_SUFFIXES
	site_packages_path = sysconfig.get_path("purelib")

	# Used to be lib but Slackware uses that for 32-bit,
	# all other distros seem to symlink lib64 to lib, so just use lib64
	libdir = "lib64"

	# Try looking in site-packages of the current environment, pwd and ../pwd
	for extension in so_extensions:
		# fmt:off
		search_paths += [
			(Path(site_packages_path)                        / (libname + extension)).resolve(),
			(Path(base_path)          / ".."                 / (libname + extension)).resolve(),
			(Path(base_path)          / ".." / ".."          / (libname + extension)).resolve(),
			# Compat with old way to store .so files
			(Path(base_path)          / ".." / ".." / libdir / (libname + ".so")).resolve(),
		]
		# fmt:on

	for path in search_paths:
		if path.exists():
			logging.debug(f"Lib {libname} found in {path!s}")
			return path

	# raise OSError(f"Can't find library '{libname}'. Searched at:\n" + "\n".join(str(p) for p in search_paths))
	return None


def get_phazor_path(pctl: PlayerCtl) -> Path:
	"""Locate the PHaZOR library in the specified priority order.

	Tries .so, .dll, .pyd, .dylib in that order and finally uses find_library as a fallback.

	:param pctl: PlayerCtl object containing installation details
	:return: Path to the library file
	:raises Exception: If no library is found
	"""
	# This is where compile-phazor.sh scripts place the dll
	base_path = Path(pctl.install_directory).parent.parent / "build"

	# Define the library name and extensions in priority order
	lib_name = "phazor-pw" if (not pctl.windows and not pctl.tauon.macos) and pctl.prefs.pipewire else "phazor"

	extensions = [".so", ".dll", ".pyd", ".dylib"]

	# Check explicitly for each file
	for ext in extensions:
		lib_path = base_path / f"lib{lib_name}{ext}"
		if lib_path.is_file():
			return lib_path

	# Fallback to find_library. Used if built using setuptools
	lib_path = find_library(lib_name)
	if lib_path:
		return Path(lib_path)

	raise Exception(f"Failed to load PHaZOR library ({lib_name})")


def phazor_exists(pctl: PlayerCtl) -> bool:
	"""Check for the existence of the PHaZOR library on the FS"""
	return get_phazor_path(pctl).exists()


_consecutive_load_failures: int = 0
_load_failure_timer: Timer = Timer()
MAX_CONSECUTIVE_LOAD_FAILURES: int = 5
LOAD_FAILURE_WINDOW: float = 3.0


def player4(tauon: Tauon) -> None:
	global _consecutive_load_failures

	def _handle_load_failure() -> bool:
		global _consecutive_load_failures
		if _load_failure_timer.get() > LOAD_FAILURE_WINDOW:
			_consecutive_load_failures = 0
		_consecutive_load_failures += 1
		_load_failure_timer.set()
		if _consecutive_load_failures >= MAX_CONSECUTIVE_LOAD_FAILURES:
			pctl.stop(run=True)
			if not gui.message_box:
				tauon.show_message(_("Multiple tracks could not be loaded"), mode="warning")
			_consecutive_load_failures = 0
			return True
		return False

	def scan_device() -> None:
		n = aud.scan_devices()
		devices = ["Default"]
		if n:
			for d in range(n):
				st = aud.get_device(d).decode()
				devices.append(st)
		prefs.phazor_devices = devices
		if prefs.phazor_device_selected not in devices:
			prefs.phazor_device_selected = devices[0]

	def pause_when_device_unavailable() -> None:
		pctl.pause_only()

	def calc_rg(track: TrackClass | None) -> float:
		if prefs.replay_gain == 0 and prefs.replay_preamp == 0:
			pctl.active_replaygain = 0
			return 0

		g = 0
		p = 1

		if track is not None:
			tg = track.replaygain_track_gain
			tp = track.replaygain_track_peak
			ag = track.replaygain_album_gain
			ap = track.replaygain_album_peak

			if prefs.replay_gain > 0:
				if prefs.replay_gain == 3 and tg is not None and ag is not None:
					gens = pctl.gen_codes.get(tauon.pl_to_id(pctl.active_playlist_playing))
					if pctl.random_mode or (gens and ("st" in gens or "rt" in gens or "r" in gens)):
						g = tg
						if tp is not None:
							p = tp
					else:
						g = ag
						if ap is not None:
							p = ap
				elif (prefs.replay_gain == 1 and tg is not None) or (
					prefs.replay_gain == 2 and ag is None and tg is not None
				):
					g = tg
					if tp is not None:
						p = tp
				elif ag is not None:
					g = ag
					if ap is not None:
						p = ap
				elif tg is not None and ag is None:
					g = tg
					if tp is not None:
						p = tp

		logging.debug("Detected ReplayGain")
		logging.debug("GAIN: " + str(g))
		logging.debug("PEAK: " + str(p))
		logging.debug("FINAL: " + str(min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)))

		if p == 0:
			logging.warning("Detected ReplayGain peak of 0")
			return 1
		pctl.active_replaygain = g
		return min(10 ** ((g + prefs.replay_preamp) / 20), 1 / p)

	def set_config(set_device: bool = False) -> None:
		aud.config_set_dev_buffer(prefs.device_buffer)
		aud.config_set_fade_duration(prefs.cross_fade_time)
		st = prefs.phazor_device_selected.encode()
		aud.config_set_dev_name(st)
		if set_device:
			aud.pause()
			aud.wait()
			aud.stop_out()
			aud.wait()
			aud.resume()
		if prefs.always_ffmpeg:
			aud.config_set_always_ffmpeg(1)
		if prefs.volume_power < 0 or prefs.volume_power > 10:
			prefs.volume_power = 2
		aud.config_set_volume_power(prefs.volume_power)
		aud.config_set_resample(prefs.avoid_resampling ^ True)
		if hasattr(aud, "config_set_stream_buffer"):
			aud.config_set_stream_buffer(prefs.stream_buffer)
		apply_eq_settings()

	def normalise_eq_bands() -> list[float]:
		try:
			bands = list(getattr(prefs, "eq", []) or [])
		except TypeError:
			bands = []
		if len(bands) < 10:
			bands.extend([0.0] * (10 - len(bands)))
		elif len(bands) > 10:
			bands = bands[:10]
		out: list[float] = []
		for v in bands:
			try:
				fv = float(v)
			except (TypeError, ValueError):
				fv = 0.0
			out.append(float(max(min(fv, 12.0), -12.0)))
		bands = out
		prefs.eq = bands
		return bands

	def apply_eq_settings() -> None:
		if not hasattr(aud, "eq_set_enable") or not hasattr(aud, "eq_set_band"):
			return
		aud.eq_set_enable(1 if prefs.use_eq else 0)
		for i, gain in enumerate(normalise_eq_bands()):
			aud.eq_set_band(i, ctypes.c_float(gain))

	def run_vis() -> None:
		if gui.turbo:  # and pctl.playing_time > 0.5:
			if gui.vis == 2:
				p_spec: list[float] = []
				aud.get_spectrum(24, bins1)
				bias = 1
				for b in list(bins1):
					p_spec.append(int(b * 1.7 * bias))
					bias += 0.04
				gui.spec = p_spec
				gui.level_update = True
				if pctl.playing_time > 0.5 and (pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM)):
					gui.update_spec = 1
			elif gui.vis == 4:
				p_spec: list[float] = []
				aud.get_spectrum(45, bins2)
				bias = 1
				for b in list(bins2):
					p_spec.append(int(b * 2.0 * bias))
					bias += 0.01
				gui.spec4_array = p_spec
				gui.level_update = True
				if pctl.playing_time > 0.5 and (pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM)):
					gui.update_spec = 1

			# Spectrogram widget: fed independently of gui.vis (own FFT buffers)
			# so it can run alongside the bar visualiser above. Raw log-spaced
			# spectrum columns, queued for SpectrogramWidget (t_custom) to
			# colourise and scroll. Frozen while paused/stopped.
			if gui.spectrogram_in_widget \
					and pctl.playing_state in (PlayingState.PLAYING, PlayingState.URL_STREAM):
				get_spectrum_spectrogram(len(bins3), bins3)
				gui.spectrogram_buffers.append(list(bins3))
				if len(gui.spectrogram_buffers) > 60:  # UI stalled; drop oldest
					del gui.spectrogram_buffers[0]
				gui.level_update = True
				if pctl.playing_time > 0.5:
					gui.update_spec = 1

	p_sync_timer = Timer()

	def track(end: bool = True) -> None:
		# run_vis() is now driven once per loop iteration in the main player
		# loop (before command processing), so commands can't starve it.

		if end and loaded_track and loaded_track.is_network and pctl.playing_time < 7 and aud.get_result() == 2:
			logging.info("STALL, RETRY")
			time.sleep(0.5)
			pctl.playerCommandReady = True
			pctl.playerCommand = "open"

		add_time = player_timer.hit()
		if add_time > 2:
			add_time = 2
		elif add_time < 0:
			add_time = 0

		# Don't advance the clock while a network stream is (re)buffering, e.g.
		# waiting on a seek's byte range — output is silent until it fills.
		# player_timer was still hit above, so the gap isn't counted later.
		if loaded_track_streamed and aud.is_buffering():
			add_time = 0

		pctl.total_playtime += add_time

		# Wait / speed up, if we are out of sync
		if p_sync_timer.get() > 1:
			real_position = aud.get_position_ms() / 1000
			if real_position and pctl.playing_time and real_position != pctl.last_real_position:
				diff = abs(real_position - pctl.playing_time)
				if 5 > diff > 0.11:  # in a CUE file real will be different that playing time
					# This assumes the first track in a CUE is > 5s
					if real_position < pctl.playing_time:
						add_time -= 5
						add_time = max(add_time, 0)
						p_sync_timer.force_set(2)  # wait for real to catch up again next clock
					else:
						add_time += 0.1
				pctl.last_real_position = real_position  # we still want to move on if playback stalled
			p_sync_timer.set()

		pctl.playing_time += add_time
		pctl.decode_time = pctl.playing_time

		if pctl.playing_time < 3 and pctl.a_time < 3:
			pctl.a_time = pctl.playing_time
		else:
			pctl.a_time += add_time

		tauon.lfm_scrobbler.update(add_time)

		if len(pctl.track_queue) > 0 and 2 > add_time > 0:
			tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)
		if end and pctl.playing_time > 1:
			pctl.test_progress()

	def chrome_start(track_id: int, enqueue: bool = False, t: int = 0) -> None:
		track = pctl.get_track(track_id)
		# if track.is_cue:
		# 	logging.error("CUE cast not supported")
		# 	return
		if track.is_network:
			if track.file_ext == "SPTY":
				logging.error("Unsupported network source for cast")
				return
			network_url, params = pctl.get_url(track)
			if params:
				req = PreparedRequest()
				req.prepare_url(network_url, params)
				network_url = req.url

			tauon.chrome.start(track.index, enqueue=enqueue, url=network_url, t=t)
		else:
			tauon.chrome.start(track.index, enqueue=enqueue, t=t)

	# fmt:off
	gui   = tauon.gui
	pctl  = tauon.pctl
	prefs = tauon.prefs
	# fmt:on
	logging.debug("Starting PHAzOR backend…")

	player_timer = Timer()
	loaded_track = None
	fade_time = 400

	aud = tauon.aud

	aud.config_set_dev_name(prefs.phazor_device_selected.encode())

	aud.init()

	aud.get_device.restype = ctypes.c_char_p

	aud.feed_raw.argtypes = (ctypes.c_int, ctypes.c_char_p)
	aud.feed_raw.restype = None
	aud.set_volume(int(pctl.player_volume))

	try:
		aud.eq_set_enable.argtypes = (ctypes.c_int,)
		aud.eq_set_enable.restype = None
		aud.eq_set_band.argtypes = (ctypes.c_int, ctypes.c_float)
		aud.eq_set_band.restype = None
	except AttributeError:
		logging.warning("PHAzOR build does not expose EQ controls")

	bins1 = (ctypes.c_float * 24)()
	bins2 = (ctypes.c_float * 45)()
	bins3 = (ctypes.c_float * gui.spectrogram_bins)()  # spectrogram widget columns
	try:
		# Doubled (4096-sample) analysis window, own C-side buffers — the
		# standard get_spectrum path the other visualisers use is untouched.
		get_spectrum_spectrogram = aud.get_spectrum_hires
	except AttributeError:  # older libphazor build
		get_spectrum_spectrogram = aud.get_spectrum

	aud.get_level_peak_l.restype = ctypes.c_float
	aud.get_level_peak_r.restype = ctypes.c_float

	active_timer = Timer()

	scan_device()
	ff_run = FFRun(tauon)

	if sys.platform == "win32":
		FUNCTYPE = WINFUNCTYPE
	else:
		FUNCTYPE = CFUNCTYPE
	start_callback = FUNCTYPE(c_int, c_char_p, c_int, c_int)(ff_run.start)
	read_callback = FUNCTYPE(c_int, c_void_p, c_int)(ff_run.read)
	close_callback = FUNCTYPE(c_void_p)(ff_run.close)
	device_unavailable_callback = FUNCTYPE(c_void_p)(pause_when_device_unavailable)
	aud.set_callbacks(start_callback, read_callback, close_callback, device_unavailable_callback)

	cachement = tauon.cachement
	feeder = StreamFeeder(tauon, aud)
	tauon.stream_feeder = feeder  # the console graph reads download stats from here
	stream_unsupported: set[int] = set()  # tracks that failed direct streaming this session
	loaded_track_streamed = False

	def set_load_net(n: int) -> None:
		if feeder.enabled:
			aud.set_load_net(n)

	# aud.config_set_samplerate(prefs.samplerate)
	aud.config_set_resample_quality(prefs.resample)

	set_config()

	stall_timer = Timer()
	wall_timer = Timer()

	chrome_update = 0
	chrome_cool_timer = Timer()
	chrome_mode = False

	while True:
		# logging.error(aud.print_status())
		time.sleep(0.016)
		if tauon.player4_state == PlayerState.PAUSED:
			time.sleep(0.05)
		if tauon.player4_state != PlayerState.STOPPED or tauon.chrome_mode:
			active_timer.set()
		if active_timer.get() > 7:
			aud.stop()
			aud.phazor_shutdown()
			break

		# Level meter
		if (tauon.player4_state in (PlayerState.PLAYING, PlayerState.URL_STREAM)) and gui.vis == 1:
			amp = aud.get_level_peak_l()
			l = amp * 12
			amp = aud.get_level_peak_r()
			r = amp * 12

			tauon.level_train.append((0, l, r))
			gui.level_update = True

		if chrome_mode:
			if tauon.chrome is None:
				logging.critical("This should not happen, tauon.chrome was None")
				return
			if pctl.playerCommandReady:
				command = pctl.playerCommand
				# logging.info(command)
				subcommand = pctl.playerSubCommand
				pctl.playerSubCommand = ""
				pctl.playerCommandReady = False

				if command == "endchrome":
					chrome_mode = False
					tauon.player4_state = PlayerState.STOPPED
					pctl.playing_time = 0
					pctl.decode_time = 0
					pctl.stop()
					continue
				if command == "open":
					target_object = pctl.target_object
					if tauon.player4_state == PlayerState.PLAYING:
						t, pid, s, d = tauon.chrome.update()
						# logging.info((t, d))
						# logging.info(d - t)

						if d and t and 1 < d - t < 5:
							# logging.info("Enqueue next chromecast")
							chrome_start(target_object.index, enqueue=True, t=pctl.start_time_target)
							chrome_cool_timer.set()
							time.sleep(d - t)
							if pctl.commit:
								pctl.advance(quiet=True, end=True)
								pctl.commit = None
							continue

					chrome_start(target_object.index, t=pctl.start_time_target)
					chrome_cool_timer.set()
					if pctl.commit:
						pctl.advance(quiet=True, end=True)
						pctl.commit = None
					tauon.player4_state = PlayerState.PLAYING
				if command == "pauseon":
					tauon.chrome.pause()
					tauon.player4_state = PlayerState.PAUSED
				if command == "pauseoff":
					tauon.chrome.play()
					tauon.player4_state = PlayerState.PLAYING
				if command == "volume":
					tauon.chrome.volume(round(pctl.player_volume / 100, 3))
					tauon.player4_state = PlayerState.PLAYING
				if command == "seek":
					tauon.chrome.seek(float(round(pctl.new_time + pctl.start_time_target, 2)))
					chrome_cool_timer.set()
					pctl.playing_time = pctl.new_time
					pctl.decode_time = pctl.playing_time
				if command == "seteq":
					apply_eq_settings()
				if command == "stop":
					tauon.player4_state = PlayerState.STOPPED
					tauon.chrome.stop()

			if tauon.player4_state == PlayerState.PLAYING:
				if chrome_update > 0.8 and chrome_cool_timer.get() > 2.5:
					t, pid, s, d = tauon.chrome.update()
					pctl.playing_time = t - pctl.start_time_target
					pctl.decode_time = t - pctl.start_time_target
					player_timer.hit()
					chrome_update = 0

				add_time = player_timer.hit()
				add_time = min(add_time, 2)
				chrome_update += add_time
				pctl.a_time += add_time
				tauon.lfm_scrobbler.update(add_time)
				pctl.total_playtime += add_time
				if len(pctl.track_queue) > 0 and 2 > add_time > 0:
					tauon.star_store.add(pctl.track_queue[pctl.queue_step], add_time)

				pctl.test_progress()

			time.sleep(0.1)
			continue

		# Feed the visualisers once per loop iteration, independent of command
		# handling. run_vis() used to run only in the play branch below (via
		# track() / the URL_STREAM path), so a burst of commands would keep
		# taking the command branch and starve it — most visibly, holding the
		# volume bar fires a "volume" command every frame, which froze the
		# spectrogram scroll. Chrome playback already returned above, so this
		# only covers local output.
		if tauon.player4_state in (PlayerState.PLAYING, PlayerState.URL_STREAM):
			run_vis()

		# Command processing
		if pctl.playerCommandReady:
			command = pctl.playerCommand
			subcommand = pctl.playerSubCommand
			pctl.playerSubCommand = ""
			pctl.playerCommandReady = False
			# logging.info(command)

			if command == "startchrome":
				aud.stop()
				if tauon.player4_state == PlayerState.PLAYING:
					chrome_start(loaded_track.index, t=pctl.playing_time)
				chrome_mode = True

			if command == "reload":
				set_config()
			if command == "set-device":
				set_config(set_device=True)

			if command == "url":
				pctl.download_time = 0
				w = 0
				if not tauon.radiobox.run_proxy:
					set_load_net(0)
					aud.start(pctl.url.encode(), 0, 0, ctypes.c_float(calc_rg(None)))
					tauon.player4_state = PlayerState.URL_STREAM
					player_timer.hit()
				else:
					while len(tauon.stream_proxy.chunks) < 200:
						time.sleep(0.1)
						w += 1
						if w > 100:
							logging.info("Taking too long!")
							tauon.stream_proxy.stop()
							pctl.playerCommand = "stop"
							pctl.playerCommandReady = True
							break
					else:
						aud.config_set_feed_samplerate(prefs.samplerate)
						set_load_net(0)
						aud.start(b"RAW FEED", 0, 0, ctypes.c_float(calc_rg(None)))
						tauon.player4_state = PlayerState.URL_STREAM
						player_timer.hit()

			if command == "open":
				if tauon.player4_state == PlayerState.PAUSED:
					aud.set_volume(int(pctl.player_volume))

				stall_timer.set()
				pctl.download_time = 0
				target_object = pctl.target_object
				if not target_object:
					logging.exception("This shouldn't happen, target_object was None")
					continue
				target_path = target_object.fullpath
				subtrack = target_object.subtrack
				aud.set_subtrack(subtrack)

				logging.info(
					f"Open - requested start was {(pctl.start_time_target + pctl.jump_time)} ({pctl.start_time_target})"
				)
				try:
					logging.info(f"Extension: {target_path.split('.')[-1]}")
				except Exception:
					logging.exception("Failed to get extension - maybe file name does not have any dots?")

				stream_url: str | None = None
				if target_object.is_network:
					if target_object.file_ext == "SPTY":
						logging.warning("This network source is no longer supported")
						target_object.found = False
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						continue

					# Prefer an already complete local copy (cache or transcode)
					instant = cachement.get_local_instant(target_object)

					if instant is None and feeder.enabled and prefs.network_stream \
							and target_object.index not in stream_unsupported:
						try:
							network_url, params = pctl.get_url(target_object)
						except Exception:
							logging.exception("Failed to resolve track URL")
							network_url = None
							params = None
						# Multi-part urls (some Tidal tracks) use the download path
						if isinstance(network_url, str) and network_url.startswith("http"):
							if params:
								req = PreparedRequest()
								req.prepare_url(network_url, params)
								stream_url = req.url
							else:
								stream_url = network_url

					if instant:
						target_path = instant
					elif stream_url:
						# Stream directly; phazor pulls the data through the feeder
						logging.info("Direct stream network track")
						target_path = stream_url
						# With the persistent cache enabled, also save the
						# stream to disk if it downloads fully (tmp_cache off
						# means the persistent cache is selected)
						if not prefs.tmp_cache:
							feeder.request_cache(target_object, stream_url)
					else:
						while True:
							status, path = cachement.get_file(target_object)

							if status in (0, 2):
								break
							# status 1: the track is uncached and downloading,
							# show buffering immediately
							if gui.buffering is False:
								gui.buffering_text = ""
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.update += 1
								tauon.wake()
							if cachement.ready == target_object and pctl.start_time_target + pctl.jump_time == 0.0:
								break
							time.sleep(0.05)
							# logging.info(status)

						gui.buffering = False
						gui.update += 1
						tauon.wake()

						if status == 2:
							logging.info("Could not locate resource")
							target_object.found = False
							pctl.playing_state = PlayingState.STOPPED
							pctl.jump_time = 0.0
							# pctl.advance(inplace=True, play=True)
							continue
						target_path = path

				elif prefs.precache:
					timer = Timer()
					timer.set()
					while True:
						status, path = cachement.get_file(target_object)
						if status in (0, 2):
							break
						if timer.get() > 0.25 and gui.buffering is False:
							gui.buffering_text = ""
							pctl.buffering_percent = 0
							gui.buffering = True
							gui.update += 1
							tauon.wake()

						time.sleep(0.05)

					gui.buffering = False
					gui.update += 1
					tauon.wake()

					if status == 2:
						target_object.found = False
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						if _handle_load_failure():
							continue
						pctl.advance(inplace=True, play=True)
						continue
					target_path = path

				if stream_url is None and not os.path.isfile(target_path):
					target_object.found = False
					if not target_object.is_network:
						pctl.playing_state = PlayingState.STOPPED
						pctl.jump_time = 0.0
						if _handle_load_failure():
							continue
						pctl.advance(inplace=True, play=True)
					continue
				_consecutive_load_failures = 0
				if not target_object.found:
					pctl.reset_missing_flags()

				# Tell the feeder which track to prefetch next, so its start
				# can be injected straight into the buffer when it plays
				if feeder.enabled and prefs.network_stream:
					try:
						nxt = pctl.advance(dry=True)
						feeder.set_prefetch(pctl.get_track(nxt) if nxt is not None else None)
					except Exception:
						logging.exception("Failed to set prefetch target")

				length: float = 0
				remain: float = 0
				position: float = 0

				if target_path and target_object and stream_url is None and target_object.length == 0 and not target_object.is_cue:
					logging.info("Track has duration of 0, scanning file")
					temp = tauon.TrackClass()
					temp.fullpath = target_path
					tauon.tag_scan(temp)
					target_object.length = temp.length
					if pctl.playing_object() is target_object:
						pctl.playing_length = target_object.length
					del temp

				if (
					tauon.player4_state == PlayerState.PLAYING
					and not pctl.start_time_target
					and not pctl.jump_time
					and loaded_track
				):
					length = aud.get_length_ms() / 1000
					position = aud.get_position_ms() / 1000
					remain = length - position

					# TODO(Martin): The GUI logger does not support multiline
					# logging.info(f"{loaded_track.title} -> {target_object.title}\nlength: {length!s}\nposition: {position!s}\nWe are {remain!s} from end")
					logging.info(f"{loaded_track.title} -> {target_object.title}")
					logging.info(f" --- length: {length!s}")
					logging.info(f" --- position: {position!s}")
					logging.info(f" --- We are {remain!s} from end")

					# Directly streamed tracks report an accurate duration from
					# the decoder; the metadata fallback is for the legacy
					# cache path where the file may still be growing
					if (loaded_track.is_network and not loaded_track_streamed) or length == 0:
						logging.warning("Phazor did not respond with a duration")
						length = loaded_track.length
						remain = length - position

				fade = 0
				error = False
				target_gapless_ready = (
					not target_object.is_network
					or stream_url is not None
					or cachement.get_file_cached_only(target_object) is not None
				)
				if (
					tauon.player4_state == PlayerState.PLAYING
					and length
					and position
					and not pctl.start_time_target
					and not pctl.jump_time
					and loaded_track
					and 0 < remain < 5.5
					and not loaded_track.is_cue
					and target_gapless_ready
					and subcommand != "now"
				):
					logging.info("Transition gapless")

					r_timer = Timer()
					r_timer.set()

					if loaded_track and loaded_track.file_ext.lower() in tauon.formats.GME:
						# GME formats dont have a physical end so we don't do gapless
						while r_timer.get() <= remain - prefs.device_buffer / 1000:
							if pctl.commit:
								track(end=False)
							time.sleep(0.016)
						aud.stop()

					set_load_net(1 if stream_url else 0)
					aud.next(
						target_path.encode(),
						int((pctl.start_time_target + pctl.jump_time) * 1000),
						ctypes.c_float(calc_rg(target_object)),
					)

					cont = False
					check_timer = Timer()
					check_timer.set()
					r_timer_saved = 0.0
					while True:
						if tauon.player4_state != PlayerState.PAUSED:
							if r_timer.get() > remain - prefs.device_buffer / 1000:
								break
						if pctl.commit and tauon.player4_state == PlayerState.PLAYING:
							track(end=False)
						time.sleep(0.016)
						if pctl.playerCommandReady and pctl.playerCommand in ("open", "stop"):
							logging.info("JANK")
							pctl.commit = None
							break
						if pctl.playerCommandReady and pctl.playerCommand == "seek":
							logging.info("Seek revert gapless")
							pctl.commit = None
							pctl.jump(loaded_track.index, jump=True)
							pctl.jump_time = pctl.new_time
							pctl.playing_time = pctl.new_time
							pctl.decode_time = pctl.new_time
							cont = True
							break
						if pctl.playerCommandReady and pctl.playerCommand == "pauseon":
							pctl.playerCommandReady = False
							tauon.player4_state = PlayerState.PAUSED
							pctl.playerCommand = ""
							aud.pause()
							r_timer_saved = r_timer.get()
						if pctl.playerCommandReady and pctl.playerCommand == "pauseoff":
							pctl.playerCommandReady = False
							tauon.player4_state = PlayerState.PLAYING
							pctl.playerCommand = ""
							aud.resume()
							player_timer.set()
							r_timer.force_set(r_timer_saved)
						if pctl.playerCommandReady and pctl.playerCommand == "volume":
							aud.ramp_volume(int(pctl.player_volume), 750)
							pctl.playerCommandReady = False
							pctl.playerCommand = ""

						if tauon.player4_state == PlayerState.PLAYING and check_timer.get() > 0.5:
							check_timer.set()
							abort_gapless = False
							if pctl.stop_mode in (1, 3):
								abort_gapless = True
							elif pctl.stop_mode in (2, 4) and pctl.commit:
								tr = pctl.playing_object()
								tr2 = pctl.get_track(pctl.commit)
								if tr and tr2 and (tr.parent_folder_path, tr.album) != (tr2.parent_folder_path, tr2.album):
									abort_gapless = True

							queue_override = pctl.force_queue and not pctl.pause_queue
							deterministic_next = queue_override or not pctl.random_mode
							if not abort_gapless and pctl.commit and deterministic_next:
								next_commit = pctl.advance(quiet=True, end=True, dry=True)
								repeat_self_commit = (
									subcommand == "repeat"
									and pctl.commit == loaded_track.index
									and target_object.index == loaded_track.index
									and not (pctl.force_queue and not pctl.pause_queue)
								)
								if next_commit != pctl.commit and not repeat_self_commit:
									abort_gapless = True

							if abort_gapless:
								resume_time = pctl.playing_time
								logging.info("Queue or stop mode revert gapless")
								pctl.abort_gapless_transition(loaded_track, resume_time)
								cont = True
								break

					if cont:
						continue
					if pctl.commit is not None:
						pctl.playing_time = 0
						pctl.decode_time = 0
						match = pctl.commit
						if match == loaded_track.index and subcommand == "repeat":
							pctl.update_change()
						else:
							pctl.advance(quiet=True, end=True)
						pt = pctl.playing_object()
						if pt and pt.index != match:
							logging.info("MISSFIRE")
							pctl.play_target()
							continue
						if pctl.playerCommandReady and pctl.playerCommand == "open":
							pctl.playerCommandReady = False
							pctl.playerCommand = ""

					loaded_track = target_object
					loaded_track_streamed = stream_url is not None
					pctl.playing_time = pctl.jump_time
				else:
					if pctl.commit and subcommand != "repeat":
						pctl.advance(quiet=True, end=True)
						pctl.commit = None
						continue

					if tauon.player4_state == PlayerState.PLAYING and prefs.use_jump_crossfade:
						fade = 1

					logging.info("Transition jump")
					# An uncached network stream isn't ready to play yet; show
					# buffering immediately and keep it until the play loop sees
					# the decoder has filled (is_buffering() clears).
					if stream_url and gui.buffering is False:
						gui.buffering = True
						gui.buffering_text = ""
						pctl.buffering_percent = 0
						gui.update += 1
						tauon.wake()
					set_load_net(1 if stream_url else 0)
					aud.start(
						target_path.encode(errors="surrogateescape"),
						int((pctl.start_time_target + pctl.jump_time) * 1000),
						fade,
						ctypes.c_float(calc_rg(target_object)),
					)
					loaded_track = target_object
					loaded_track_streamed = stream_url is not None
					pctl.playing_time = pctl.jump_time
					if pctl.jump_time:
						while aud.get_result() == 0:
							if stream_url and pctl.playerCommandReady and pctl.playerCommand in ("stop", "open"):
								break
							time.sleep(0.016)
							run_vis()
						aud.set_position_ms(int(pctl.jump_time * 1000))

					# Restart track is failed to load (for some network tracks) (broken with gapless loading)
					stream_fallback = False
					buff_timer = Timer()
					while True:
						r = aud.get_result()
						if r == 1:
							break
						if r == 2:
							if stream_url:
								# Direct streaming didn't work out; fall back
								# to the legacy download-to-cache path
								logging.warning("Direct stream failed, falling back to download")
								stream_unsupported.add(target_object.index)
								pctl.playerCommand = "open"
								pctl.playerCommandReady = True
								stream_fallback = True
								break
							if loaded_track.is_network:
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.buffering_text = ""

								# while dm.request(loaded_track, whole=True) == "wait":
								# 	time.sleep(0.05)
								# 	if pctl.playerCommandReady:
								# 		break
								logging.info("Retry start file")
								set_load_net(0)
								aud.start(
									target_path.encode(),
									int((pctl.start_time_target + pctl.jump_time) * 1000),
									fade,
									ctypes.c_float(calc_rg(target_object)),
								)
								gui.buffering = False
								player_timer.set()
								break
							aud.stop()
							if not gui.message_box:
								tauon.show_message(_("Error loading track"), mode="warning")
							error = True
							break
						if stream_url:
							# Keep commands responsive if the network stalls during load
							if pctl.playerCommandReady and pctl.playerCommand in ("stop", "open"):
								logging.info("Command received while stream loading, cancel load")
								aud.stop()
								error = True
								break
							if buff_timer.get() > 0.25 and gui.buffering is False:
								gui.buffering_text = ""
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.update += 1
								tauon.wake()
						time.sleep(0.016)
						run_vis()

					# For a successful direct stream, the decoder has loaded but
					# the PCM buffer may still be filling; leave the buffering
					# flag for the play loop to clear once is_buffering() reports
					# ready. Clear now for non-streamed tracks (ready already) or
					# if the load errored/aborted (the play loop won't run).
					if gui.buffering and (error or not loaded_track_streamed):
						gui.buffering = False
						gui.update += 1
						tauon.wake()
					if stream_fallback:
						continue

					tauon.player4_state = PlayerState.PLAYING
					if error:
						tauon.player4_state = PlayerState.STOPPED

				player_timer.set()
				pctl.jump_time = 0.0
				if loaded_track.length == 0 or loaded_track.file_ext.lower() in tauon.formats.MOD:
					i = 0
					t = 0
					while t == 0:
						time.sleep(0.3)
						t = aud.get_length_ms() / 1000
						i += 1
						if i > 9:
							break
					loaded_track.length = t
					if loaded_track.length != 0:
						pctl.playing_length = loaded_track.length
						gui.pl_update += 1

				pctl.commit = None
				stall_timer.set()
				wall_timer.force_set(3)

			if command == "seek":
				if tauon.player4_state != PlayerState.STOPPED:
					if loaded_track.is_network and loaded_track_streamed:
						# Direct streams seek natively; the decoder requests
						# the byte range it needs through the feeder. Show the
						# buffering state right away — the play loop freezes the
						# clock and clears this once the new position has filled.
						aud.seek(int((pctl.new_time + pctl.start_time_target) * 1000), prefs.pa_fast_seek)
						if gui.buffering is False:
							gui.buffering = True
							gui.buffering_text = ""
							gui.update += 1
							tauon.wake()
					elif loaded_track.is_network:  # and loaded_track.fullpath.endswith(".ogg"):
						timer = Timer()
						timer.set()
						i = 0
						while True:
							status, path = cachement.get_file(loaded_track)
							if status == 1:
								per = (pctl.new_time / loaded_track.length) * 100
								if per < 1:
									break
								if pctl.buffering_percent - per > 5:
									break
							if status in (0, 2):
								break
							if timer.get() > 0.25 and gui.buffering is False:
								gui.buffering_text = ""
								pctl.buffering_percent = 0
								gui.buffering = True
								gui.update += 1
								tauon.wake()
							if i * 0.05 > 2:
								aud.pause()
							if pctl.playerCommandReady:
								break

							time.sleep(0.05)
							i += 1

						gui.buffering = False
						gui.update += 1
						tauon.wake()
						if pctl.playerCommandReady:
							continue

						if status == 2:
							loaded_track.found = False
							pctl.playing_state = PlayingState.STOPPED
							pctl.jump_time = 0.0
							pctl.stop()
							continue

						# The vorbis decoder doesn't like appended files
						set_load_net(0)
						aud.start(
							path.encode(),
							int(pctl.new_time + pctl.start_time_target) * 1000,
							0,
							ctypes.c_float(calc_rg(loaded_track)),
						)
						while aud.get_result() == 0:
							time.sleep(0.01)
					else:
						aud.seek(int((pctl.new_time + pctl.start_time_target) * 1000), prefs.pa_fast_seek)

					pctl.playing_time = pctl.new_time

				pctl.decode_time = pctl.playing_time
				wall_timer.set()
				player_timer.set()

			if command == "volume":
				aud.ramp_volume(int(pctl.player_volume), 750)

			if command == "seteq":
				apply_eq_settings()

			if command == "runstop":
				length = aud.get_length_ms() / 1000
				position = aud.get_position_ms() / 1000
				remain = length - position
				# logging.info("length: " + str(length))
				# logging.info("position: " + str(position))
				# logging.info("We are %s from end" % str(remain))
				time.sleep(3)
				command = "stop"

			if command == "stop":
				if prefs.use_pause_fade and tauon.player4_state != PlayerState.URL_STREAM:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)

				tauon.player4_state = PlayerState.STOPPED
				pctl.playing_time = 0
				aud.stop()
				time.sleep(0.1)
				aud.set_volume(int(pctl.player_volume))

				# Clear any buffering indicator; the play loop will not run now
				if gui.buffering:
					gui.buffering = False
					gui.update += 1
					tauon.wake()

				if subcommand == "return":
					pctl.playerSubCommand = "stopped"
					# pctl.playerCommandReady = True

			if command == "pauseon":
				if prefs.use_pause_fade:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)
				aud.pause()
				tauon.player4_state = PlayerState.PAUSED

			if command == "pauseoff":
				if tauon.player4_state == PlayerState.STOPPED:
					t = pctl.playing_time
					pctl.play_target()
					pctl.jump_time = t
					pctl.playing_time = t
					pctl.decode_time = t
				else:
					if prefs.use_pause_fade:
						if pctl.player_volume > 5:
							speed = fade_time / (int(pctl.player_volume) / 100)
						else:
							speed = fade_time / (5 / 100)

						aud.ramp_volume(int(pctl.player_volume), int(speed))
					aud.resume()
					player_timer.set()
					stall_timer.set()
					tauon.player4_state = PlayerState.PLAYING

			if command == "unload":
				if prefs.use_pause_fade:
					if pctl.player_volume > 5:
						speed = fade_time / (int(pctl.player_volume) / 100)
					else:
						speed = fade_time / (5 / 100)

					aud.ramp_volume(0, int(speed))
					time.sleep((fade_time + 100) / 1000)

				aud.stop()
				aud.phazor_shutdown()

				if cachement.audio_cache.exists():
					shutil.rmtree(cachement.audio_cache)

				pctl.playerCommand = "done"
				pctl.playerCommandReady = True
				break
		else:
			if tauon.player4_state == PlayerState.URL_STREAM:
				pctl.radio_progress()
				# run_vis() runs once per iteration at the top of the loop now.

				add_time = player_timer.hit()
				pctl.playing_time += add_time
				pctl.decode_time = pctl.playing_time

				buffering = aud.is_buffering()
				if gui.buffering != buffering:
					gui.buffering = buffering
					gui.update += 1

			if tauon.player4_state == PlayerState.PLAYING:
				if loaded_track_streamed:
					# Surface buffering state if the network can't keep up
					buffering = aud.is_buffering()
					if gui.buffering != buffering:
						gui.buffering = buffering
						gui.buffering_text = ""
						gui.update += 1
						tauon.wake()
				track()



"""Tauon Music Box - Theme reading module"""

# Copyright © 2015-2022, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

import base64
import io
import json
import logging
import os
from typing import TYPE_CHECKING, Protocol

import sdl3
from PIL import Image

from tauon.t_modules.t_extra import ColourRGBA, rgb_add_hls, test_lumi

if TYPE_CHECKING:
	from pathlib import Path

	from tauon.t_modules.t_draw import TDraw
	from tauon.t_modules.t_main import ColoursClass, Directories, PlayerCtl, Tauon
	from tauon.t_modules.t_prefs import Prefs


class GetThemesFn(Protocol):
	def __call__(self, *, dirs: Directories, deco: bool) -> list[str] | dict[str, str]: ...


THEME_SERIALISABLE_COLOURS: tuple[tuple[str, str], ...] = (
	("window frame", "window_frame"),
	("gallery highlight", "gallery_highlight"),
	("index playing", "index_playing"),
	("time playing", "time_text"),
	("artist playing", "artist_playing"),
	("track album", "album_text"),
	("album playing", "album_playing"),
	("top panel", "top_panel_background"),
	("corner button off", "corner_button"),
	("corner button on", "corner_button_active"),
	("menu button normal", "status_text_normal"),
	("menu button hover", "status_text_over"),
	("queue panel", "queue_background"),
	("side panel", "side_panel_background"),
	("lyrics panel", "lyrics_panel_background"),
	("gallery background", "gallery_background"),
	("tracklist panel", "playlist_panel_background"),
	("track line", "title_text"),
	("track missing", "playlist_text_missing"),
	("playing highlight", "row_playing_highlight"),
	("track time", "bar_time"),
	("fav line", "star_line"),
	("folder title", "folder_title"),
	("folder line", "folder_line"),
	("buttons off", "media_buttons_off"),
	("buttons over", "media_buttons_over"),
	("buttons active", "media_buttons_active"),
	("playing time", "time_playing"),
	("track index", "index_text"),
	("track playing", "title_playing"),
	("select highlight", "row_select_highlight"),
	("track artist", "artist_text"),
	("tab active text", "tab_text_active"),
	("tab text", "tab_text"),
	("tab background", "tab_background"),
	("tab over", "tab_highlight"),
	("tab active background", "tab_background_active"),
	("title info", "side_bar_line1"),
	("extra info", "side_bar_line2"),
	("bottom title", "bar_title_text"),
	("scroll bar", "scroll_colour"),
	("seek bar", "seek_bar_fill"),
	("seek bg", "seek_bar_background"),
	("volume bar", "volume_bar_fill"),
	("volume bg", "volume_bar_background"),
	("mode off", "mode_button_off"),
	("mode over", "mode_button_over"),
	("mode on", "mode_button_active"),
	("art border", "art_box"),
	("tb line", "tb_line"),
	("music vis", "vis_colour"),
	("menu background", "menu_background"),
	("menu text", "menu_text"),
	("menu disable", "menu_text_disabled"),
	("menu icons", "menu_icons"),
	("menu highlight", "menu_highlight_background"),
	("menu border", "menu_tab"),
	("lyrics text", "lyrics"),
	("active lyric", "active_lyric"),
	("bottom panel", "bottom_panel_colour"),
	("mini bg", "mini_mode_background"),
	("mini border", "mini_mode_border"),
	("mini text 1", "mini_mode_text_1"),
	("mini text 2", "mini_mode_text_2"),
	("playlist background", "playlist_box_background"),
	("box background", "box_background"),
	("box border", "box_border"),
	("box text border", "box_text_border"),
	("box text label", "box_text_label"),
	("box title text", "box_title_text"),
	("box text normal", "box_text"),
	("box sub text", "box_sub_text"),
	("box input text", "box_input_text"),
	("box button text highlight", "box_button_text_highlight"),
	("box button text normal", "box_button_text"),
	("box button background normal", "box_button_background"),
	("box button background highlight", "box_button_background_highlight"),
	("box button border", "box_check_border"),
	("window buttons background", "window_buttons_bg"),
	("window buttons on", "window_buttons_bg_over"),
	("window buttons icon off", "window_button_icon_off"),
	("window buttons icon over", "window_buttons_icon_over"),
	("window button x on", "window_button_x_on"),
	("window button x off", "window_button_x_off"),
	("column bar background", "column_bar_background"),
	("artist bio background", "artist_bio_background"),
	("artist bio text", "artist_bio_text"),
	("visualizer bg", "vis_bg"),
	("link text", "link_text"),
	("star line playing", "star_line_playing"),
	("message box bg", "message_box_bg"),
	("message box text", "message_box_text"),
	("queue drag indicator", "queue_drag_indicator_colour"),
	("pulse colour", "pulse_colour"),
	("queue card background", "queue_card_background"),
	("column grip", "column_grip"),
	("column bar text", "column_bar_text"),
	("box thumb background", "box_thumb_background"),
)


def get_colour_from_line(cline: str) -> ColourRGBA:
	colour: list[int] = [-1, -1, -1, -1]
	colour_str: list[str] = ["", "", "", ""]
	mode = 0

	is_hex = False
	if "," not in cline[:8]:
		if cline[:6].isalnum():
			is_hex = True
		if cline[0] == "#" and cline[1:7].isalnum() and cline[7].isspace() and cline[8].isspace():
			is_hex = True
		if cline[0] == "#" and cline[1:9].isalnum() and cline[9].isspace() and cline[10].isspace():
			is_hex = True

	if is_hex:
		# hex mode
		cline = cline.removeprefix("0x")
		cline = cline.removeprefix("#")
		ll = cline
		a = 255
		try:
			r = int(ll[0] + ll[1], 16)
			g = int(ll[2] + ll[3], 16)
			b = int(ll[4] + ll[5], 16)
			if ll[6].isalnum() and ll[7].isalnum():
				a = int(ll[6] + ll[7], 16)
		except ValueError:
			logging.error(f"Error parsing theme line {cline}, we will ignore it")
			return ColourRGBA(255, 255, 255, 255)
		except Exception:
			logging.exception(f"Unknown error parsing theme line {cline}")
			return ColourRGBA(255, 255, 255, 255)
		return ColourRGBA(r, g, b, a)
	# rgb mode
	for i in cline:
		if i.isdigit():
			colour_str[mode] += i
		elif i == "," and mode < 3:
			mode += 1
		# Stop parsing after RGB/RGBA values so we don't absorb digits in labels such as "mini text 2"
		elif (i.isspace() and mode >= 2 and colour_str[mode]) or (mode >= 2 and colour_str[mode]):
			break

	# Convert str list to int list
	for b in range(len(colour_str)):
		if colour_str[b] == "":
			colour[b] = 255
		else:
			colour[b] = int(colour_str[b])

	if (
		(colour[0] < 0 or colour[0] > 255) or
		(colour[1] < 0 or colour[1] > 255) or
		(colour[2] < 0 or colour[2] > 255) or
		(colour[3] < 0 or colour[3] > 255)
	):
		logging.error(f"Got nonsense colour value parsed from {cline}, we will ignore it")
		return ColourRGBA(255, 255, 255, 255)

	return ColourRGBA(colour[0], colour[1], colour[2], colour[3])


def save_theme(colours: ColoursClass, path: Path) -> None:
	lines = ["# Tauon user theme\n"]

	if colours.deco:
		lines.append(f"deco={colours.deco}\n")
	if colours.lm:
		lines.append("light-mode\n")

	for key, attr in THEME_SERIALISABLE_COLOURS:
		colour = getattr(colours, attr, None)
		if colour is None:
			continue
		lines.append(f"{colour.r},{colour.g},{colour.b},{colour.a}    {key}\n")

	for key, value in sorted(colours.column_colours.items()):
		lines.append(f"{value.r},{value.g},{value.b},{value.a}    column-{key.lower().replace(' ', '-')}\n")
	for key, value in sorted(colours.column_colours_playing.items()):
		lines.append(f"{value.r},{value.g},{value.b},{value.a}    column+{key.lower().replace(' ', '-')}\n")

	path.parent.mkdir(parents=True, exist_ok=True)
	path.write_text("".join(lines), encoding="utf-8")


def load_theme(colours: ColoursClass, path: Path) -> None:
	with path.open(encoding="utf-8") as f:
		content = f.readlines()

		for p in content:
			p = p.strip()
			if p.startswith("#"):
				continue
			if "# " in p:
				p = p.split("# ")[0]
			if not p:
				continue
			if p[0] == "#" and ("-" in p[:7] or " " in p[:7] or "\t" in p[:7]):
				continue
			if p.startswith("deco="):
				colours.deco = p.split("=", 1)[1].strip()
			elif "light-mode" in p:
				colours.light_mode()
			elif "window frame" in p:
				colours.window_frame = get_colour_from_line(p)
			elif "gallery highlight" in p:
				colours.gallery_highlight = get_colour_from_line(p)
			elif "index playing" in p:
				colours.index_playing = get_colour_from_line(p)
			elif "time playing" in p:
				colours.time_text = get_colour_from_line(p)
			elif "artist playing" in p:
				colours.artist_playing = get_colour_from_line(p)
			elif "album line" in p or "track album" in p:  # Bad name
				colours.album_text = get_colour_from_line(p)
			elif "album playing" in p:
				colours.album_playing = get_colour_from_line(p)
			elif "top panel" in p or "player background" in p:
				colours.top_panel_background = get_colour_from_line(p)

				if test_lumi(colours.bottom_panel_colour) < 0.2:
					colours.corner_icon = ColourRGBA(0, 0, 0, 60)
				elif test_lumi(colours.bottom_panel_colour) < 0.8:
					colours.corner_icon = ColourRGBA(40, 40, 40, 255)
				else:
					colours.corner_icon = ColourRGBA(255, 255, 255, 30)

				if not colours.lm:
					colours.corner_button = rgb_add_hls(colours.top_panel_background, 0, 0.18, 0)
			elif "status text" in p:
				status_text_temp = get_colour_from_line(p)
				colours.status_text_over = status_text_temp
				colours.status_text_normal = rgb_add_hls(status_text_temp, 0, 0.30, -0.15)

			elif "corner button off" in p:
				colours.corner_button = get_colour_from_line(p)
			elif "corner button on" in p:
				colours.corner_button_active = get_colour_from_line(p)
			elif "menu button normal" in p:
				colours.status_text_normal = get_colour_from_line(p)
				status_text_color_defined = 1
			elif "menu button hover" in p:
				colours.status_text_over = get_colour_from_line(p)
				status_text_color_defined = 1
			elif "queue panel" in p:
				colours.queue_background = get_colour_from_line(p)
			elif "side panel" in p:
				colours.side_panel_background = get_colour_from_line(p)
				colours.playlist_box_background = colours.side_panel_background
			elif "lyrics panel" in p:
				colours.lyrics_panel_background = get_colour_from_line(p)
				lyrics_panel_color_defined = 1
			elif "gallery background" in p:
				colours.gallery_background = get_colour_from_line(p)
			elif "playlist panel" in p or "tracklist panel" in p:  # bad name
				colours.playlist_panel_background = get_colour_from_line(p)
			elif "track line" in p:
				colours.title_text = get_colour_from_line(p)
			elif "track missing" in p:
				colours.playlist_text_missing = get_colour_from_line(p)
			elif "playing highlight" in p:
				colours.row_playing_highlight = get_colour_from_line(p)
			elif "track time" in p:
				colours.bar_time = get_colour_from_line(p)
			elif "fav line" in p:
				colours.star_line = get_colour_from_line(p)
			elif "folder title" in p:
				colours.folder_title = get_colour_from_line(p)
			elif "folder line" in p:
				colours.folder_line = get_colour_from_line(p)
			elif "buttons off" in p:
				colours.media_buttons_off = get_colour_from_line(p)
			elif "buttons over" in p:
				colours.media_buttons_over = get_colour_from_line(p)
			elif "buttons active" in p:
				colours.media_buttons_active = get_colour_from_line(p)
			elif "playing time" in p:
				colours.time_playing = get_colour_from_line(p)
			elif "track index" in p:
				colours.index_text = get_colour_from_line(p)
			elif "track playing" in p:
				colours.title_playing = get_colour_from_line(p)
			elif "select highlight" in p:
				colours.row_select_highlight = get_colour_from_line(p)
			elif "track artist" in p:
				colours.artist_text = get_colour_from_line(p)
			elif "tab active line" in p or "tab active text" in p:  # bad name
				colours.tab_text_active = get_colour_from_line(p)
			elif "tab line" in p or "tab text" in p:  # bad name
				colours.tab_text = get_colour_from_line(p)
			elif "tab background" in p:
				colours.tab_background = get_colour_from_line(p)
			elif "tab over" in p:
				colours.tab_highlight = get_colour_from_line(p)
			elif "tab active background" in p:
				colours.tab_background_active = get_colour_from_line(p)
			elif "title info" in p:
				colours.side_bar_line1 = get_colour_from_line(p)
			elif "extra info" in p:
				colours.side_bar_line2 = get_colour_from_line(p)
			elif "bottom title" in p:
				colours.bar_title_text = get_colour_from_line(p)
			elif "scroll bar" in p:
				colours.scroll_colour = get_colour_from_line(p)
			elif "seek bar" in p:
				colours.seek_bar_fill = get_colour_from_line(p)
			elif "seek bg" in p:
				colours.seek_bar_background = get_colour_from_line(p)
			elif "volume bar" in p:
				colours.volume_bar_fill = get_colour_from_line(p)
			elif "volume bg" in p:
				colours.volume_bar_background = get_colour_from_line(p)
			elif "mode off" in p:
				colours.mode_button_off = get_colour_from_line(p)
			elif "mode over" in p:
				colours.mode_button_over = get_colour_from_line(p)
			elif "mode on" in p:
				colours.mode_button_active = get_colour_from_line(p)
			elif "art border" in p:
				colours.art_box = get_colour_from_line(p)
			elif "tb line" in p:
				colours.tb_line = get_colour_from_line(p)
			elif "music vis" in p:
				colours.vis_colour = get_colour_from_line(p)
			elif "menu background" in p or "menu bg" in p:
				colours.menu_background = get_colour_from_line(p)
			elif "menu text" in p:
				colours.menu_text = get_colour_from_line(p)
			elif "menu disable" in p:
				colours.menu_text_disabled = get_colour_from_line(p)
			elif "menu icons" in p:
				colours.menu_icons = get_colour_from_line(p)
			elif "menu highlight" in p:
				colours.menu_highlight_background = get_colour_from_line(p)
			elif "menu border" in p:
				colours.menu_tab = get_colour_from_line(p)
			elif "lyrics showcase" in p or "lyrics text" in p:
				colours.lyrics = get_colour_from_line(p)
			elif "active lyric" in p:
				colours.active_lyric = get_colour_from_line(p)
			elif "bottom panel" in p:
				colours.bottom_panel_colour = get_colour_from_line(p)
				# colours.menu_background = colours.bottom_panel_colour
			elif "mini bg" in p:
				colours.mini_mode_background = get_colour_from_line(p)
			elif "mini border" in p:
				colours.mini_mode_border = get_colour_from_line(p)
			elif "mini text 1" in p:
				colours.mini_mode_text_1 = get_colour_from_line(p)
			elif "mini text 2" in p:
				colours.mini_mode_text_2 = get_colour_from_line(p)
			elif "column-" in p:
				key = p[p.find("column-") + 7 :].replace("-", " ").lower().title().rstrip()
				value = get_colour_from_line(p)
				colours.column_colours[key] = value
			elif "column+" in p:
				key = p[p.find("column+") + 7 :].replace("-", " ").lower().title().rstrip()
				value = get_colour_from_line(p)
				colours.column_colours_playing[key] = value
			elif "playlist box bg" in p or "playlist background" in p:  # bad name
				colours.playlist_box_background = get_colour_from_line(p)

			elif "box background" in p:
				colours.box_background = get_colour_from_line(p)
			elif "box border" in p:
				colours.box_border = get_colour_from_line(p)
			elif "box text border" in p:
				colours.box_text_border = get_colour_from_line(p)
			elif "box text label" in p:
				colours.box_text_label = get_colour_from_line(p)

			elif "box title text" in p:
				colours.box_title_text = get_colour_from_line(p)
			elif "box text normal" in p:
				colours.box_text = get_colour_from_line(p)
			elif "box sub text" in p:
				colours.box_sub_text = get_colour_from_line(p)
			elif "box input text" in p:
				colours.box_input_text = get_colour_from_line(p)

			elif "box button text highlight" in p:
				colours.box_button_text_highlight = get_colour_from_line(p)
			elif "box button text normal" in p:
				colours.box_button_text = get_colour_from_line(p)
			elif "box button background normal" in p:
				colours.box_button_background = get_colour_from_line(p)
			elif "box button background highlight" in p:
				colours.box_button_background_highlight = get_colour_from_line(p)
			elif "box button border" in p:
				colours.box_check_border = get_colour_from_line(p)

			elif "window buttons background" in p:
				colours.window_buttons_bg = get_colour_from_line(p)
			elif "window buttons on" in p:
				colours.window_buttons_bg_over = get_colour_from_line(p)
			elif "window buttons icon off" in p:
				colours.window_button_icon_off = get_colour_from_line(p)
				colours.window_button_x_off = colours.window_button_icon_off
			elif "window buttons icon over" in p:
				colours.window_buttons_icon_over = get_colour_from_line(p)
				colours.window_button_x_on = colours.window_buttons_icon_over
			elif "window button x on" in p:
				colours.window_button_x_on = get_colour_from_line(p)
			elif "window button x off" in p:
				colours.window_button_x_off = get_colour_from_line(p)
			elif "column bar background" in p:
				colours.column_bar_background = get_colour_from_line(p)

			elif "artist bio background" in p:
				colours.artist_bio_background = get_colour_from_line(p)
			elif "artist bio text" in p:
				colours.artist_bio_text = get_colour_from_line(p)

			elif "visualizer bg" in p:
				colours.vis_bg = get_colour_from_line(p)
			elif "link text" in p:
				colours.link_text = get_colour_from_line(p)
			elif "star line playing" in p:
				colours.star_line_playing = get_colour_from_line(p)
			elif "message box bg" in p:
				colours.message_box_bg = get_colour_from_line(p)
			elif "message box text" in p:
				colours.message_box_text = get_colour_from_line(p)
			elif "queue drag indicator" in p:
				colours.queue_drag_indicator_colour = get_colour_from_line(p)
			elif "pulse colour" in p:
				colours.pulse_colour = get_colour_from_line(p)
			elif "queue card background" in p:
				colours.queue_card_background = get_colour_from_line(p)
			elif "column grip" in p:
				colours.column_grip = get_colour_from_line(p)
			elif "column bar text" in p:
				colours.column_bar_text = get_colour_from_line(p)
			elif "box thumb background" in p:
				colours.box_thumb_background = get_colour_from_line(p)
			# elif "panel button off" in p:
			# 	colours.corner_button = get_colour_from_line(p)
			# elif "panel button on" in p:
			# 	colours.corner_button_active = get_colour_from_line(p)
	colours.post_config()
	if colours.lm:
		colours.light_mode()


class Drawable:
	def __init__(self) -> None:
		self.location = 1
		self.x = 0
		self.y = 0
		self.w = 100
		self.h = 100
		self.rect = None
		self.texture = None


class Deco:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon: Tauon = tauon
		self.renderer = None
		self.pctl: PlayerCtl = tauon.pctl
		self.prefs: Prefs = tauon.prefs
		self.drawables: list[Drawable] = []
		self.get_themes: GetThemesFn

	def unload(self) -> None:
		for item in self.drawables:
			sdl3.SDL_DestroyTexture(item.texture)
		self.drawables.clear()

	def load(self, name: str) -> None:
		self.unload()
		if not name:
			return

		decos = self.get_themes(deco=True, dirs=self.tauon.dirs)

		if name not in decos:
			logging.error("Missing deco file")
			return

		target = decos[name]
		with open(target) as f:
			j = json.load(f)

		if "images" not in j:
			return
		images = j["images"]
		if not images:
			return
		item = images[0]
		if item.get("location") != 1:
			return
		x = int(item.get("x-margin", 0))
		y = int(item.get("y-margin", 0))
		opacity = int(item.get("opacity", 100))
		logical_h = int(item.get("logical-height", 400))
		file = item.get("file")
		if file:
			path = os.path.join(os.path.dirname(target), file)
			im = Image.open(path)
		else:
			s = item.get("image-data")
			if not s:
				return
			b = io.BytesIO(base64.b64decode(s))
			im = Image.open(b)

		w, h = im.size
		scale = self.tauon.gui.scale
		if not abs(h - (logical_h * scale)) < 10:
			new_h = round(logical_h * scale)
			ratio = w / h
			new_w = round(new_h * ratio)
			im = im.resize((new_w, new_h), Image.Resampling.LANCZOS)
			w, h = new_w, new_h

		g = io.BytesIO()
		g.seek(0)
		im.save(g, "PNG")
		g.seek(0)
		s_image = self.tauon.ddt.load_image(g)
		texture = sdl3.SDL_CreateTextureFromSurface(self.renderer, s_image)
		sdl3.SDL_SetTextureAlphaMod(texture, opacity)
		sdl3.SDL_DestroySurface(s_image)
		sdl_rect = sdl3.SDL_FRect(0, 0, w, h)

		drawable = Drawable()
		drawable.x = x
		drawable.y = y
		drawable.w = w
		drawable.h = h
		drawable.rect = sdl_rect
		drawable.texture = texture
		self.drawables.append(drawable)

	def draw(self, ddt: TDraw, x: int, y: int, pretty_text: bool = False) -> None:
		if self.drawables:
			d = self.drawables[0]
			d.rect.x = round(x - int(d.w + round(d.x * self.tauon.gui.scale)))
			d.rect.y = round(y - int(d.h + round(d.y * self.tauon.gui.scale)))
			sdl3.SDL_RenderTexture(self.renderer, d.texture, None, d.rect)

			if pretty_text:
				ddt.pretty_rect = (d.rect.x, d.rect.y, d.rect.w, d.rect.h)
				ddt.alpha_bg = True



"""Tauon Music Box - SVG Module"""

# Copyright © 2019-2020, Taiko2k captain(dot)gxj(at)gmail.com

#     This file is part of Tauon Music Box.
#
#     Tauon Music Box is free software: you can redistribute it and/or modify
#     it under the terms of the GNU General Public License as published by
#     the Free Software Foundation, either version 3 of the License, or
#     (at your option) any later version.
#
#     Tauon Music Box is distributed in the hope that it will be useful,
#     but WITHOUT ANY WARRANTY; without even the implied warranty of
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#     GNU Lesser General Public License for more details.
#
#     You should have received a copy of the GNU General Public License
#     along with Tauon Music Box.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

import cairo
from gi import require_version

require_version("Rsvg", "2.0")
from gi.repository import Rsvg


def render_icons(source_directory: str, output_directory: str, scale: int) -> None:
	"""Render SVG files to PNG"""
	targets = []
	# Verify svg files exist
	for file in os.listdir(source_directory):
		if file.endswith(".svg") and os.path.isfile(os.path.join(source_directory, file)):
			targets.append(file)

	if not os.path.exists(output_directory):
		os.makedirs(output_directory)

	# Render
	for file in targets:
		name = os.path.splitext(file)[0]
		in_path = os.path.join(source_directory, file)
		out_path = os.path.join(output_directory, name + ".png")

		handle = Rsvg.Handle()
		svg = handle.new_from_file(in_path)
		if svg is None:
			logging.error(f"Couldn't load {in_path}")
			continue
		unscaled_width = svg.props.width
		unscaled_height = svg.props.height

		width = unscaled_width * scale
		height = unscaled_height * scale

		svg_surface = cairo.SVGSurface(None, width, height)
		svg_context = cairo.Context(svg_surface)

		viewport = Rsvg.Rectangle()
		viewport.x = 0
		viewport.y = 0
		viewport.width = width
		viewport.height = height

		svg.render_document(svg_context, viewport)

		svg_surface.write_to_png(out_path)



from __future__ import annotations

import logging
import socket
import threading
from typing import TYPE_CHECKING

import pychromecast
import zeroconf
from pychromecast.controllers.media import BaseMediaPlayer, MediaController, MediaStatus

from tauon.t_modules.t_extra import shooter

if TYPE_CHECKING:
	from pychromecast import Chromecast
	from pychromecast.discovery import CastBrowser

	from tauon.t_modules.t_main import Tauon


DISCOVERY_TIMEOUT = 5.0
STYLED_RECEIVER_APP_ID = "2F76715B"


def get_ip() -> str:
	s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
	s.settimeout(0)
	try:
		# doesn't even have to be reachable
		s.connect(("10.255.255.255", 1))
		ipv4 = s.getsockname()[0]
	except Exception:
		logging.exception("Failed to get socket name.")
		ipv4 = "127.0.0.1"
	finally:
		s.close()
	return ipv4


class StyledMediaController(MediaController):
	def __init__(self, app_id: str) -> None:
		BaseMediaPlayer.__init__(self, supporting_app_id=app_id, app_must_match=False)
		self.media_session_id = 0
		self.status = MediaStatus()
		self.session_active_event = threading.Event()
		self._status_listeners = []


class Chrome:
	def __init__(self, tauon: Tauon) -> None:
		self.tauon: Tauon = tauon
		self.services: list[list[str]] = []
		self.active: bool = False
		self.cast: Chromecast | None = None
		self.save_vol: float = 100
		self.ip: str = ""
		self.browser: CastBrowser | None = None
		self.media_controller: StyledMediaController | None = None

	def discover_services(self, timeout: float = DISCOVERY_TIMEOUT) -> list[list[str]]:
		zconf = zeroconf.Zeroconf()
		try:
			browser = pychromecast.discovery.CastBrowser(
				pychromecast.discovery.SimpleCastListener(),
				zconf,
			)
		except Exception:
			zconf.close()
			raise
		try:
			browser.start_discovery()
			threading.Event().wait(timeout)
		finally:
			browser.stop_discovery()
		return sorted(
			[
				[str(device.uuid), str(device.friendly_name)]
				for device in browser.devices.values()
			],
			key=lambda service: service[1].casefold(),
		)

	def rescan(self) -> None:
		logging.info("Scanning for chromecasts...")

		if True:  # not self.services:
			try:
				# self.tauon.show_message(self.tauon.strings.scan_chrome)
				services = self.discover_services()
				menu = self.tauon.chrome_menu
				if menu is None:
					logging.critical("menu was None, this should not happen!")
					return
				MenuItem = self.tauon.MenuItem
				chrome_submenu_index = len(menu.subs) - 1
				if chrome_submenu_index < 0:
					logging.critical("Chromecast submenu was missing, this should not happen!")
					return

				self.services = services
				menu.subs[chrome_submenu_index].clear()
				for item in self.services:
					menu.add_to_sub(
						chrome_submenu_index,
						MenuItem(
							self.tauon.strings.cast_to % item[1],
							self.three,
							pass_ref=True,
							args=item,
						),
					)
				menu.add_to_sub(
					chrome_submenu_index,
					MenuItem(self.tauon.strings.stop_cast, self.end, show_test=lambda x: self.active),
				)
			except Exception:
				logging.exception("Failed to get chromecasts")
				raise

	def three(self, _, item: tuple) -> None:
		shooter(self.four, [item])

	def four(self, item: list) -> None:
		if self.active:
			self.end()
		self.tauon.start_remote()
		ccs, browser = pychromecast.get_listed_chromecasts(friendly_names=[item[1]], discovery_timeout=3.0)
		self.browser = browser
		if not ccs:
			logging.error("No Chromecast found for selected device")
			return
		self.cast = ccs[0]
		self.cast.wait()
		self.save_vol = self.tauon.pctl.player_volume
		volume = self.cast.status.volume_level if self.cast.status else 0
		if not self.cast.status:
			logging.critical("self.cast.status was None, this should not happen!")
		self.tauon.pctl.player_volume = min(volume * 100, 100)
		self.ip = get_ip()
		self.media_controller = StyledMediaController(STYLED_RECEIVER_APP_ID)
		self.cast.register_handler(self.media_controller)

		self.tauon.chrome_mode = True
		self.active = True
		self.tauon.gui.update += 1
		self.tauon.pctl.playerCommand = "startchrome"
		self.tauon.pctl.playerCommandReady = True
		self.tauon.thread_manager.ready_playback()

	def update(self) -> tuple:
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return (0.0, None, None, 0.0)
		self.media_controller.update_status()
		status = self.media_controller.status
		if status is None:
			logging.critical("self.media_controller.status was None, this should not happen!")
			return (0.0, None, None, 0.0)
		custom_data = status.media_custom_data if isinstance(status.media_custom_data, dict) else {}
		return (
			status.current_time,
			custom_data.get("id"),
			status.player_state,
			status.duration,
		)

	def start(self, track_id: int, enqueue: bool = False, t: int = 0, url: str | None = None) -> None:
		if self.cast is None:
			logging.critical("self.cast was None, this should not happen!")
			return
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return
		self.cast.wait()
		self.cast.start_app(STYLED_RECEIVER_APP_ID)
		tr = self.tauon.pctl.get_track(track_id)
		n = 0
		try:
			n = int(tr.track_number)
		except Exception:
			logging.exception("Failed to get track number")
		d = {
			"metadataType": 3,
			"albumName": tr.album,
			"title": tr.title,
			"albumArtist": tr.album_artist,
			"artist": tr.artist,
			"trackNumber": n,
			"images": [{"url": f"http://{self.ip}:7814/api1/pic/medium/{track_id}"}],
			"releaseDate": tr.date,
		}
		m = {
			"duration": round(float(tr.length), 1),
			"customData": {"id": str(tr.index)},
		}

		if url is None:
			url = f"http://{self.ip}:7814/api1/file/{track_id}"
		else:
			url = url.replace("localhost", self.ip)
			url = url.replace("127.0.0.1", self.ip)

		self.media_controller.play_media(
			url, "audio/mpeg", media_info=m, metadata=d, current_time=t, enqueue=enqueue
		)

	def stop(self) -> None:
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return
		self.media_controller.stop()

	def play(self) -> None:
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return
		self.media_controller.play()

	def pause(self) -> None:
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return
		self.media_controller.pause()

	def seek(self, position: float) -> None:
		if self.media_controller is None:
			logging.critical("self.media_controller was None, this should not happen!")
			return
		self.media_controller.seek(position)

	def volume(self, decimal: int) -> None:
		if self.cast is None:
			logging.critical("self.cast was None, this should not happen!")
			return
		self.cast.set_volume(decimal)

	def end(self) -> None:
		self.tauon.pctl.playerCommand = "endchrome"
		self.tauon.pctl.playerCommandReady = True
		if self.active:
			if self.media_controller:
				self.media_controller.stop()
			self.active = False
		self.tauon.chrome_mode = False
		self.tauon.pctl.player_volume = self.save_vol
		self.media_controller = None



#!/usr/bin/env python3
"""Tauon Music Box"""

# Copyright © 2015-2025, Taiko2k captain(dot)gxj(at)gmail.com

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import ctypes
import ctypes.util
import logging
import os
import pickle
import subprocess
import sys
from ctypes import c_int, pointer
from pathlib import Path

from gi.repository import GLib

install_directory = Path(__file__).resolve().parent
sys.path.insert(0, str(install_directory.parent))

from tauon.t_modules.t_bootstrap import Holder  # noqa: E402
from tauon.t_modules.t_logging import CustomLoggingFormatter, LogHistoryHandler  # noqa: E402

pyinstaller_mode = bool(
	hasattr(sys, "_MEIPASS") or getattr(sys, "frozen", False) or install_directory.name.endswith("_internal")
)

log = LogHistoryHandler()
formatter = logging.Formatter("[%(levelname)s] %(message)s")
log.setFormatter(formatter)

# DEBUG+ to file and std_err
logging.basicConfig(
	level=logging.DEBUG,
	handlers=[
		logging.StreamHandler(),
		log,
		# logging.FileHandler('/tmp/tauon.log'),
	],
)
logging.getLogger().handlers[0].setFormatter(CustomLoggingFormatter())

# https://docs.python.org/3/library/warnings.html
logging.captureWarnings(capture=True)
if not sys.warnoptions:
	import warnings

	warnings.simplefilter("default")
	os.environ["PYTHONWARNINGS"] = "default"  # Also affect subprocesses

if sys.platform != "win32":
	import fcntl

n_version = "11.0.0"  # Should also be bumped in pyproject.toml, extra/*.appdata.xml
t_version = "v" + n_version
t_title = "Tauon"
t_id = "tauonmb"
if str(install_directory).startswith("/app/") or sys.platform == "darwin":
	# Its Flatpak or macOS
	t_id = "com.github.taiko2k.tauonmb"
t_agent = f"{t_title}/" + n_version
t_creator = "Taiko2k"
t_copyright = "Copyright 2015-2026 Taiko2k"
t_url = "https://tauonmusicbox.rocks/"
t_type = "mediaplayer"

logging.info(f"{t_title} {t_version}")
logging.info(f"{t_copyright} captain.gxj@gmail.com\n")

logging.info(f"Started with arguments: {sys.argv}")


def set_sdl_app_metadata() -> None:
	"""Populate SDL app metadata before SDL_Init()."""
	if not sdl3.SDL_SetAppMetadata(
		t_title.encode("utf-8"),
		n_version.encode("utf-8"),
		t_id.encode("utf-8"),
	):
		logging.warning("Failed to set SDL app metadata")

	metadata_properties = (
		(sdl3.SDL_PROP_APP_METADATA_NAME_STRING, t_title.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_VERSION_STRING, n_version.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_IDENTIFIER_STRING, t_id.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_CREATOR_STRING, t_creator.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_COPYRIGHT_STRING, t_copyright.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_URL_STRING, t_url.encode("utf-8")),
		(sdl3.SDL_PROP_APP_METADATA_TYPE_STRING, t_type.encode("utf-8")),
	)
	for key, value in metadata_properties:
		if not sdl3.SDL_SetAppMetadataProperty(key, value):
			logging.warning(f"Failed to set SDL app metadata property: {key!r}")


def open_discord() -> None:
	webbrowser.open("https://discord.gg/v4EmhES")


def open_github() -> None:
	webbrowser.open("https://github.com/Taiko2k/Tauon/issues")


def open_crash_log(path: Path) -> None:
	try:
		if sys.platform == "win32":
			os.startfile(path)  # type: ignore[attr-defined]
		elif sys.platform == "darwin":
			subprocess.Popen(["open", str(path)])
		else:
			subprocess.Popen(["xdg-open", str(path)])
	except Exception:
		webbrowser.open(path.as_uri())


def main() -> None:
	"""Launch Tauon by means of importing t_main.py"""
	from tauon.t_modules.t_main import main as t_main

	t_main(holder)


def transfer_args_and_exit() -> None:
	"""Early arg processing"""
	import urllib.request

	base = "http://localhost:7813/"

	if len(sys.argv) <= 1:
		url = base + "raise/"
		urllib.request.urlopen(url)

	for item in sys.argv:
		if (
			not item.endswith(".py")
			and not item.startswith("-")
			and not item.endswith("exe")
			and (item.startswith("file://") or Path(item).exists())
		):
			import base64

			url = base + "open/" + base64.urlsafe_b64encode(item.encode()).decode()
			urllib.request.urlopen(url)
		if item == "--play-pause":
			url = base + "playpause/"
			urllib.request.urlopen(url)
		if item == "--play":
			url = base + "play/"
			urllib.request.urlopen(url)
		if item == "--pause":
			url = base + "pause/"
			urllib.request.urlopen(url)
		if item == "--stop":
			url = base + "stop/"
			urllib.request.urlopen(url)
		if item == "--next":
			url = base + "next/"
			urllib.request.urlopen(url)
		if item == "--previous":
			url = base + "previous/"
			urllib.request.urlopen(url)
		if item == "--shuffle":
			url = base + "shuffle/"
			urllib.request.urlopen(url)
		if item == "--repeat":
			url = base + "repeat/"
			urllib.request.urlopen(url)
		if item == "--reload-theme":
			url = base + "reloadtheme/"
			urllib.request.urlopen(url)

	sys.exit()


if "--no-start" in sys.argv:
	transfer_args_and_exit()

# If we're installed, use home data locations
install_mode = bool(
	str(install_directory).startswith(("/opt/", "/usr/", "/app/", "/snap/", "/nix/store/"))
	or sys.platform in ("darwin", "win32")
)

# Assume that it's a classic Linux install, use standard paths
if str(install_directory).startswith("/usr/") and Path("/usr/share/TauonMusicBox").is_dir():
	install_directory = Path("/usr/share/TauonMusicBox")

if (install_directory / "portable").is_file():
	install_mode = False

# Handle regular install, running from a directory and finally a portable install, usually a venv
if install_mode:
	# logging.info("Running in installed mode")
	user_directory = Path(GLib.get_user_data_dir()) / "TauonMusicBox"
elif install_directory.parent.name == "src":
	# logging.info("Running in portable mode from cloned dir")
	user_directory = install_directory.parent.parent / "user-data"
else:
	# logging.info("Running in portable mode")
	user_directory = install_directory / "user-data"

debug = bool((user_directory / "debug").is_file())

# INFO+ to std_err
# TODO(Martin): This hereabout section is wonk, setting INFO on streamhandler removes formatting for DEBUG
logging.getLogger().handlers[0].setLevel(logging.DEBUG if debug else logging.INFO)

asset_directory = install_directory / "assets"

if not user_directory.is_dir():
	user_directory.mkdir(parents=True)

if debug:
	file_handler = logging.FileHandler(user_directory / "tauon.log")
	file_handler.setLevel(logging.DEBUG)
	file_handler.setFormatter(formatter)
	logging.getLogger().addHandler(file_handler)
	logging.info(f"Debug mode enabled - saving log to {user_directory / 'tauon.log'}")

fp = None
dev_mode = (install_directory / ".dev").is_file()
if dev_mode:
	logging.warning("Dev mode, ignoring single instancing")
elif sys.platform != "win32":
	pid_file = user_directory / "program.pid"
	fp = pid_file.open("w")
	try:
		fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
	except OSError:
		logging.exception("Another Tauon instance is already running")
		# TODO(Martin): Silent crash
		transfer_args_and_exit()
else:
	if sys.platform == "win32":
		pid_file = user_directory / "program.pid"
		try:
			if pid_file.is_file():
				pid_file.unlink()
			fp = pid_file.open("w")
		except OSError:
			logging.exception("Another Tauon instance is already running")
			# TODO(Martin): Silent crash
			transfer_args_and_exit()
	if pyinstaller_mode:
		os.environ["FONTCONFIG_PATH"] = str(install_directory / "etc" / "fonts")  # "C:\\msys64\\mingw64\\etc\\fonts"

phone = False
d = os.environ.get("XDG_CURRENT_DESKTOP")
if d in ["GNOME:Phosh"]:
	os.environ["SDL_VIDEODRIVER"] = "wayland"
	phone = True

os.environ["SDL_VIDEO_WAYLAND_ALLOW_LIBDECOR"] = "0"  # emergency crash workaround

if pyinstaller_mode:  # and sys.platform == 'darwin':
	os.environ["SDL_BINARY_PATH"] = str(install_directory)

fs_mode = False
if os.environ.get("GAMESCOPE_WAYLAND_DISPLAY") is not None:
	fs_mode = True
	logging.info("Running in GAMESCOPE MODE")

if os.environ.get("XDG_SESSION_TYPE") and os.environ.get("XDG_SESSION_TYPE") == "wayland":
	# Force Wayland, as SDL3 otherwise requires the compositor to support specific protocols
	# and defaults to X11 - https://github.com/libsdl-org/SDL/pull/9383
	# We should be able to remove this in 2026+
	os.environ["SDL_VIDEODRIVER"] = "wayland"
if Path(user_directory / "x11").exists():
	logging.debug("Forcing X11 due to user prefs")
	os.environ["SDL_VIDEODRIVER"] = "x11"

# Pysdl3 doesn't seem to find system sdl3. REF: https://github.com/Aermoss/PySDL3/issues/35 pysdl
# As a workaround we try find it ourselves
# find_library() can return both a filename "libSDL3.so.0" or a full path "/opt/homebrew/lib/libSDL3.dylib"
sdl_dll_name = ctypes.util.find_library("SDL3")
if sdl_dll_name:
	sdl_dll_name = str(Path(sdl_dll_name).name)
	candidates = [
		"./" + sdl_dll_name,
		os.path.join("/app/lib", sdl_dll_name),
		os.path.join("/usr/lib64", sdl_dll_name),
		os.path.join("/lib64", sdl_dll_name),
		os.path.join("/usr/local/lib64", sdl_dll_name),
		os.path.join("/usr/lib", sdl_dll_name),
		os.path.join("/lib", sdl_dll_name),
		os.path.join("/usr/local/lib", sdl_dll_name),
	]
else:
	candidates = []
sdl_library_path = next((p for p in candidates if p and os.path.exists(p)), None)
if sdl_library_path:
	os.environ["SDL_BINARY_PATH"] = os.path.dirname(sdl_library_path)

os.environ["SDL_FIND_BINARIES"] = "1"
if pyinstaller_mode:
	os.environ["SDL_FIND_BINARIES"]            = "0" # Disable binary searching,                   "1"        by default.

os.environ["SDL_DOC_GENERATOR"]            = "0" # Disable doc generation                       "1"        by default.
os.environ["SDL_DISABLE_METADATA"]         = "1" # Disable metadata method,                     "0"        by default.
os.environ["SDL_CHECK_VERSION"]            = "0" # Disable version checking,                    "1"        by default.
os.environ["SDL_CHECK_BINARY_VERSION"]     = "0" # Disable binary version checking,             "1"        by default.
os.environ["SDL_IGNORE_MISSING_FUNCTIONS"] = "1" # Disable missing function warnings,           "0"        by default.

import sdl3  # noqa: E402

# Test the first SetHint to catch if we loaded SDL3 correctly
sethint_result = sdl3.SDL_SetHint(sdl3.SDL_HINT_VIDEO_ALLOW_SCREENSAVER, b"1")
if sethint_result is None:
	logging.error(
		"Failed to run SetHint, probably due to https://github.com/Aermoss/PySDL3/issues/35, will try a workaround"
	)
	sys.exit(1)

set_sdl_app_metadata()

sdl3.SDL_SetHint(sdl3.SDL_HINT_MOUSE_FOCUS_CLICKTHROUGH, b"1")
sdl3.SDL_SetHint(sdl3.SDL_HINT_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR, b"0")
if sys.platform == "darwin" and hasattr(sdl3, "SDL_HINT_MAC_SCROLL_MOMENTUM"):
	sdl3.SDL_SetHint(sdl3.SDL_HINT_MAC_SCROLL_MOMENTUM, b"1")
# sdl3.SDL_SetHint(sdl3.SDL_HINT_APP_ID, t_id.encode("utf-8"))
# sdl3.SDL_SetHint(sdl3.SDL_HINT_APP_NAME, t_title.encode("utf-8"))

draw_border = True
w = 1120
h = 600
if phone:
	w = 720
	h = 1800
window_default_size: tuple[int, int] = (w, h)
window_size: list[int] = [w, h]
logical_size: list[int] = [w, h]
window_opacity = 1
scale = 1
if sys.platform == "darwin":
	scale = 2
if phone:
	scale = 1.3

maximized = False
old_window_position: tuple[int, int] | None = None

window_p = user_directory / "window.p"
if window_p.is_file() and not fs_mode:
	try:
		state_file = window_p.open("rb")
		save = pickle.load(state_file)
		state_file.close()

		draw_border = save[0]
		window_size = save[1]
		w, h = save[1]
		if 100 < w < 10000 and 100 < h < 5000:
			logical_size[0], logical_size[1] = w, h
		window_opacity = save[2]
		scale = save[3]
		maximized = save[4]
		old_window_position = save[5]
		del save

	except Exception:
		logging.exception("Corrupted window state file?! Please restart app!")
		window_p.unlink()
		sys.exit(1)
else:
	logging.info("No window state file")


##  Maybe this is needed any more?
# if d == "GNOME": #and os.environ.get("XDG_SESSION_TYPE") and os.environ.get("XDG_SESSION_TYPE") == "wayland":
# 	try:
# 		import gi.repository
# 		# TODO(Martin): Bump to 4.0 - https://github.com/Taiko2k/Tauon/issues/1316
# 		gi.require_version("Gtk", "3.0")
# 		from gi.repository import Gtk
#
#
# 		gtk_settings = Gtk.Settings().get_default()
# 		xtheme = gtk_settings.get_property("gtk-cursor-theme-name")
# 		xsize = gtk_settings.get_property("gtk-cursor-theme-size")
# 		os.environ["XCURSOR_THEME"] = xtheme
# 		os.environ["XCURSOR_SIZE"] = str(xsize)
# 	except Exception:
# 		logging.exception("Failed to set cursor")

sdl3.SDL_Init(sdl3.SDL_INIT_VIDEO | sdl3.SDL_INIT_EVENTS)

err = sdl3.SDL_GetError()
if err and "GLX" in err.decode():
	logging.error(f"SDL init error: {err.decode()}")
	sdl3.SDL_ShowSimpleMessageBox(
		sdl3.SDL_MESSAGEBOX_ERROR,
		b"Tauon Music Box failed to start :(",
		b"Error: " + err + b".\n If you're using Flatpak, try running `$ flatpak update`",
		None,
	)
	sys.exit(1)

window_title = t_title
window_title = window_title.encode("utf-8")

flags = sdl3.SDL_WINDOW_RESIZABLE | sdl3.SDL_WINDOW_TRANSPARENT | sdl3.SDL_WINDOW_HIGH_PIXEL_DENSITY

if draw_border and not fs_mode:
	flags |= sdl3.SDL_WINDOW_BORDERLESS

if fs_mode:
	flags |= sdl3.SDL_WINDOW_FULLSCREEN

if old_window_position is None:
	o_x = sdl3.SDL_WINDOWPOS_UNDEFINED
	o_y = sdl3.SDL_WINDOWPOS_UNDEFINED
else:
	o_x = old_window_position[0]
	o_y = old_window_position[1]

if "--tray" in sys.argv:
	flags |= sdl3.SDL_WINDOW_HIDDEN


t_window = sdl3.SDL_CreateWindow(  # TODO(Taiko): use SDL_CreateWindowAndRenderer()
	window_title,
	# o_x, o_y,
	logical_size[0],
	logical_size[1],
	flags,
)

if not t_window:
	logging.error("ERROR CREATING WINDOW!")
	logging.error(f"Title: {window_title}")
	logging.error(f"X: {o_x}")
	logging.error(f"Y: {o_y}")
	logging.error(f"Size 0: {logical_size[0]}")
	logging.error(f"Size 1: {logical_size[1]}")
	logging.error(f"Flags: {flags}")
	sdl_err = sdl3.SDL_GetError()
	logging.error(f"SDL Error: {sdl_err}")
	if sdl_err and sdl_err.decode() == "x11 not available":
		x11_path = user_directory / "x11"
		if x11_path.exists():
			logging.critical(
				"Disabled Xwayland preference as X11 was not found - Known issue if on Flatpak - https://github.com/Taiko2k/Tauon/issues/1034"
			)
			x11_path.unlink()
			# TODO(Martin): This does not seem to work on SDL3, it attempts to relaunch under x11 again
			# os.environ["SDL_VIDEODRIVER"] = "wayland"
			# t_window = sdl3.SDL_CreateWindow(
			# 	window_title,
			# 	o_x, o_y,
			# 	logical_size[0], logical_size[1],
			# 	flags)
			# if not t_window:
			# 	logging.error(f"Failed to create Wayland fallback window - SDL Error: {sdl3.SDL_GetError()}")
			# 	sys.exit(1)
			sys.exit(1)
		else:
			logging.critical(f"Failed to find {x11_path} but got 'x11 not available' error, hm?")
			sys.exit(1)

if maximized:
	sdl3.SDL_MaximizeWindow(t_window)

drivers: list[bytes] = []
i = 0
while True:
	x = sdl3.SDL_GetRenderDriver(i)
	i += 1
	if x is None:
		break
	drivers.append(x)

logging.debug(f"SDL available drivers: {drivers}")
logging.debug(f"PATH that will be used for ffmpeg/ffprobe and similar: {os.environ.get('PATH')}")

driver = None
if b"opengl" in drivers:
	driver = b"opengl"

renderer = sdl3.SDL_CreateRenderer(t_window, driver)  # sdl3.SDL_RENDERER_PRESENTVSYNC
if not renderer and driver == b"opengl":
	renderer_error = sdl3.SDL_GetError()
	logging.warning(f"Failed to create OpenGL renderer: {renderer_error}")
	logging.warning("Trying SDL default renderer")
	sdl3.SDL_ClearError()
	renderer = sdl3.SDL_CreateRenderer(t_window, None)

if not renderer:
	logging.error("ERROR CREATING RENDERER!")
	logging.error(f"SDL Error: {sdl3.SDL_GetError()}")
	sys.exit(1)

sdl3.SDL_SetRenderDrawBlendMode(renderer, sdl3.SDL_BLENDMODE_BLEND)
sdl3.SDL_SetWindowOpacity(t_window, window_opacity)

sdl3.SDL_SetRenderDrawColor(renderer, 0, 0, 0, 0)
sdl3.SDL_RenderClear(renderer)

logging.info(f"SDL window system: {sdl3.SDL_GetCurrentVideoDriver().decode()}")

# Vsync is not trusted for frame pacing (visualisers self-pace to the display
# refresh rate, see Tauon.frame_pace) — enabling it just aligns frames to vblank.
sdl3.SDL_SetRenderVSync(
	renderer, 1
)  # 1 == enable vsync, 0 == disable, -1 == late swap tearing, 2 == second vertical refresh

i_x = pointer(c_int(0))
i_y = pointer(c_int(0))
sdl3.SDL_GetWindowSizeInPixels(t_window, i_x, i_y)
window_size[0] = i_x.contents.value
window_size[1] = i_y.contents.value

sdl3.SDL_GetWindowSize(t_window, i_x, i_y)
logical_size[0] = i_x.contents.value
logical_size[1] = i_y.contents.value

# Loading screen: a full window grid of interconnected isometric wireframe boxes
def draw_loading_screen() -> None:
	box_w = round(44 * scale)  # half-width of a box top face
	box_r = box_w // 2         # half-height of the top face (2:1 isometric)
	box_d = round(35 * scale)  # vertical edge length

	def polyline(points: list[tuple[float, float]]) -> None:
		array = (sdl3.SDL_FPoint * len(points))()
		for i, (x, y) in enumerate(points):
			array[i].x = float(x)
			array[i].y = float(y)
		sdl3.SDL_RenderLines(renderer, array, len(points))

	sdl3.SDL_SetRenderDrawColor(renderer, 7, 7, 7, 255)
	sdl3.SDL_RenderFillRect(renderer, None)

	sdl3.SDL_SetRenderDrawColor(renderer, 120, 134, 150, 35)
	# Boxes tile seamlessly: each box's bottom front edges and right vertical
	# coincide with its neighbours' top faces and left vertical, so per box only
	# the top face and the west + south verticals are drawn to avoid doubling up
	cy = -box_r
	row = 0
	while cy < window_size[1] + box_r + box_d:
		cx = -box_w * 2 + (box_w if row % 2 else 0)
		while cx < window_size[0] + box_w * 2:
			n = (cx, cy - box_r)
			e = (cx + box_w, cy)
			s = (cx, cy + box_r)
			w = (cx - box_w, cy)
			polyline([n, e, s, w, n])
			polyline([w, (w[0], w[1] + box_d)])
			polyline([s, (s[0], s[1] + box_d)])
			cx += box_w * 2
		cy += box_r + box_d
		row += 1

	sdl3.SDL_RenderPresent(renderer)


draw_loading_screen()

holder = Holder(
	t_window=t_window,
	renderer=renderer,
	logical_size=logical_size,
	window_size=window_size,
	window_default_size=window_default_size,
	scale=scale,
	maximized=maximized,
	transfer_args_and_exit=transfer_args_and_exit,
	draw_border=draw_border,
	window_opacity=window_opacity,
	old_window_position=old_window_position,
	install_directory=install_directory,
	user_directory=user_directory,
	pyinstaller_mode=pyinstaller_mode,
	phone=phone,
	window_title=window_title,
	fs_mode=fs_mode,
	t_title=t_title,
	n_version=n_version,
	t_version=t_version,
	t_id=t_id,
	t_agent=t_agent,
	dev_mode=dev_mode,
	instance_lock=fp,
	log=log,
)

del flags

if __name__ == "__main__":
	try:
		main()
	except Exception as e:
		crash_logger = logging.getLogger("crash_logger")
		crash_logger.setLevel(logging.DEBUG)
		crash_log_path = user_directory / "tauon-crash.log"
		file_handler = logging.FileHandler(crash_log_path)
		crash_logger.addHandler(file_handler)
		crash_logger.handlers[0].setFormatter(CustomLoggingFormatter(color=False))
		error_message = f"Oops, looks like Tauon crashed.\n\nPlease report a bug over at GitHub or Discord.\n\nCrash log was saved to\n{crash_log_path}"
		crash_logger.exception(error_message)
		import ctypes
		import webbrowser

		BUTTON_ID_GITHUB = 1
		BUTTON_ID_DISCORD = 2
		BUTTON_ID_OPEN_LOG = 3
		BUTTON_ID_QUIT = 4

		# Create buttons
		button_array = (sdl3.SDL_MessageBoxButtonData * 4)(
			sdl3.SDL_MessageBoxButtonData(0, BUTTON_ID_GITHUB, b"GitHub"),
			sdl3.SDL_MessageBoxButtonData(0, BUTTON_ID_DISCORD, b"Discord"),
			sdl3.SDL_MessageBoxButtonData(0, BUTTON_ID_OPEN_LOG, b"Open Crash Log"),
			sdl3.SDL_MessageBoxButtonData(sdl3.SDL_MESSAGEBOX_BUTTON_ESCAPEKEY_DEFAULT, BUTTON_ID_QUIT, b"Quit"),
		)
		messageboxdata = sdl3.SDL_MessageBoxData(
			flags=sdl3.SDL_MESSAGEBOX_ERROR,
			window=None,
			title=b"Tauon Music Box crashed :(",
			message=str.encode(f"{error_message}\n\nShortlog:\n{e}"),
			numbuttons=4,
			buttons=button_array,
			color_scheme=None,
		)
		buttonid = ctypes.c_int(-1)

		# Show the message box
		success = sdl3.SDL_ShowMessageBox(ctypes.byref(messageboxdata), ctypes.byref(buttonid))

		if not success:
			logging.error(f"SDL_ShowMessageBox failed: {sdl3.SDL_GetError().decode()}")
		elif buttonid.value == BUTTON_ID_DISCORD:
			open_discord()
		elif buttonid.value == BUTTON_ID_GITHUB:
			open_github()
		elif buttonid.value == BUTTON_ID_OPEN_LOG:
			open_crash_log(crash_log_path)
		elif buttonid.value == BUTTON_ID_QUIT:
			pass
